import numpy.array_api as nxp

from cubed.core import CoreArray, map_blocks


def astype(x, dtype, /, *, copy=True):
    if not copy and dtype == x.dtype:
        return x
    return map_blocks(_astype, x, dtype=dtype, astype_dtype=dtype)


def _astype(a, astype_dtype):
    return a.astype(astype_dtype)


def can_cast(from_, to, /):
    if isinstance(from_, CoreArray):
        from_ = from_.dtype
    return nxp.can_cast(from_, to)


def finfo(type, /):
    return nxp.finfo(type)


def iinfo(type, /):
    return nxp.iinfo(type)


def result_type(*arrays_and_dtypes):
    # Use numpy.array_api promotion rules (stricter than numpy)
    return nxp.result_type(
        *(a.dtype if isinstance(a, CoreArray) else a for a in arrays_and_dtypes)
    )



from dataclasses import dataclass
from operator import mul
from typing import Optional, TypeVar

import numpy as np
from dask.array.core import normalize_chunks
from toolz import map, reduce

from cubed.runtime.pipeline import already_computed
from cubed.runtime.types import Executor
from cubed.utils import chunk_memory

from .plan import arrays_to_plan

sym_counter = 0


def gensym(name="array"):
    """Generate a name with an incrementing counter"""
    global sym_counter
    sym_counter += 1
    return f"{name}-{sym_counter:03}"


# type representing either a CoreArray or a public-facing Array
T_ChunkedArray = TypeVar("T_ChunkedArray", bound="CoreArray")


class CoreArray:
    """Chunked array backed by Zarr storage.

    This class has the basic array attributes (``dtype``, ``shape``, ``chunks``, etc).
    The other array methods and attributes are provided in a subclass.
    """

    def __init__(self, name, zarray, spec, plan):
        self.name = name
        self.zarray = zarray
        self._shape = zarray.shape
        self._dtype = zarray.dtype
        self._chunks = normalize_chunks(
            zarray.chunks, shape=self.shape, dtype=self.dtype
        )
        # if no spec is supplied, use a default with local temp dir, a modest amount of memory (100MB),
        # and a conservative amount of reserved memory (200MB)
        self.spec = spec or Spec(None, max_mem=100_000_000, reserved_mem=200_000_000)
        self.plan = plan

    @property
    def chunkmem(self):
        """Amount of memory in bytes that a single chunk uses."""
        return chunk_memory(self.dtype, self.chunksize)

    @property
    def chunksize(self):
        """A tuple indicating the chunk size of each corresponding array dimension."""
        return tuple(max(c) for c in self.chunks)

    @property
    def chunks(self):
        """A tuple containing a sequence of block sizes for each corresponding array dimension."""
        return self._chunks

    @property
    def dtype(self):
        """Data type of the array elements."""
        return self._dtype

    @property
    def ndim(self):
        """Number of array dimensions (axes)."""
        return len(self.shape)

    @property
    def numblocks(self):
        """A tuple indicating the number of blocks (chunks) in each corresponding array dimension."""
        return tuple(map(len, self.chunks))

    @property
    def npartitions(self):
        """Number of chunks in the array."""
        return reduce(mul, self.numblocks, 1)

    @property
    def shape(self):
        """Array dimensions."""
        return self._shape

    @property
    def size(self):
        """Number of elements in the array."""
        return reduce(mul, self.shape, 1)

    def _read_stored(self):
        # Only works if the array has been computed
        if self.size > 0:
            # read back from zarr
            return self.zarray[...]
        else:
            # this case fails for zarr, so just return an empty array of the correct shape
            return np.empty(self.shape, dtype=self.dtype)

    def compute(
        self,
        *,
        executor=None,
        callbacks=None,
        optimize_graph=True,
        **kwargs,
    ):
        """Compute this array, and any arrays that it depends on."""
        result = compute(
            self,
            executor=executor,
            callbacks=callbacks,
            optimize_graph=optimize_graph,
            **kwargs,
        )
        if result:
            return result[0]

    def rechunk(self: T_ChunkedArray, chunks) -> T_ChunkedArray:
        """Change the chunking of this array without changing its shape or data.

        Parameters
        ----------
        chunks : tuple
            The desired chunks of the array after rechunking.

        Returns
        -------
        cubed.CoreArray
            An array with the desired chunks.
        """
        from cubed.core.ops import rechunk

        return rechunk(self, chunks)

    def visualize(self, filename="cubed", format=None, optimize_graph=True):
        """Produce a visualization of the computation graph for this array.

        Parameters
        ----------
        filename : str
            The name of the file to write to disk. If the provided ``filename``
            doesn't include an extension, '.svg' will be used by default.
        format : {'png', 'pdf', 'dot', 'svg', 'jpeg', 'jpg'}, optional
            Format in which to write output file.  Default is 'svg'.
        optimize_graph : bool, optional
            If True, the graph is optimized before rendering.  Otherwise,
            the graph is displayed as is. Default is True.

        Returns
        -------
        IPython.display.SVG, or None
            An IPython SVG image if IPython can be imported (for rendering
            in a notebook), otherwise None.
        """
        return visualize(
            self, filename=filename, format=format, optimize_graph=optimize_graph
        )

    def __getitem__(self: T_ChunkedArray, key, /) -> T_ChunkedArray:
        from cubed.core.ops import index

        return index(self, key)

    def __setitem__(self, key, value):
        if isinstance(value, CoreArray) and value.ndim != 0:
            raise NotImplementedError(
                "Calling __setitem__ on an array with more than 0 dimensions is not supported."
            )

        nodes = {n: d for (n, d) in self.plan.dag.nodes(data=True)}
        if not already_computed(nodes[self.name]):
            raise NotImplementedError(
                "Calling __setitem__ on an array that has not been computed is not supported."
            )

        self.zarray.__setitem__(key, value)

    def __repr__(self):
        return f"cubed.core.CoreArray<{self.name}, shape={self.shape}, dtype={self.dtype}, chunks={self.chunks}>"


@dataclass
class Spec:
    """Specification of resources available to run a computation."""

    work_dir: Optional[str] = None
    """The directory path (specified as an fsspec URL) used for storing intermediate data."""

    max_mem: Optional[int] = None
    """The maximum memory available to a worker for data use for the computation, in bytes."""

    reserved_mem: Optional[int] = None
    """The memory reserved on a worker for non-data use, in bytes.

    The sum of ``max_mem`` and ``reserved_mem`` should not exceed the total memory of the worker.

    See ``cubed.measure_reserved_memory``.
    """

    executor: Executor = None
    """The default executor for running computations."""

    storage_options: dict = None  # type: ignore[assignment]
    """Storage options to be passed to fsspec"""


class Callback:
    """Object to receive callback events during array computation."""

    def on_compute_start(self, dag):
        """Called when the computation is about to start.

        Parameters
        ----------
        dag : networkx.MultiDiGraph
            The computation DAG.
        """
        pass  # pragma: no cover

    def on_compute_end(self, dag):
        """Called when the computation has finished.

        Parameters
        ----------
        dag : networkx.MultiDiGraph
            The computation DAG.
        """
        pass  # pragma: no cover

    def on_task_end(self, event):
        """Called when the a task ends.

        Parameters
        ----------
        event : TaskEndEvent
            Information about the task execution.
        """
        pass  # pragma: no cover


@dataclass
class TaskEndEvent:
    """Callback information about a completed task (or tasks)."""

    array_name: str
    """Name of the array that the task is for."""

    num_tasks: int = 1
    """Number of tasks that this event applies to (default 1)."""

    task_create_tstamp: Optional[float] = None
    """Timestamp of when the task was created by the client."""

    function_start_tstamp: Optional[float] = None
    """Timestamp of when the function started executing on the remote worker."""

    function_end_tstamp: Optional[float] = None
    """Timestamp of when the function finished executing on the remote worker."""

    task_result_tstamp: Optional[float] = None
    """Timestamp of when the result of the task was received by the client."""

    peak_memory_start: Optional[int] = None
    """Peak memory usage on the remote worker before the function starts executing."""

    peak_memory_end: Optional[int] = None
    """Peak memory usage on the remote worker after the function finishes executing."""


def check_array_specs(arrays):
    specs = [a.spec for a in arrays if hasattr(a, "spec")]
    if not all(s == specs[0] for s in specs):
        raise ValueError(
            f"Arrays must have same spec in single computation. Specs: {specs}"
        )
    return arrays[0].spec


def compute(
    *arrays,
    executor=None,
    callbacks=None,
    optimize_graph=True,
    **kwargs,
):
    """Compute multiple arrays at once."""
    plan = arrays_to_plan(*arrays)  # guarantees all arrays have same spec
    if executor is None:
        executor = arrays[0].spec.executor
        if executor is None:
            from cubed.runtime.executors.python import PythonDagExecutor

            executor = PythonDagExecutor()

    _return_in_memory_array = kwargs.pop("_return_in_memory_array", True)
    plan.execute(
        executor=executor,
        callbacks=callbacks,
        optimize_graph=optimize_graph,
        array_names=[a.name for a in arrays],
        **kwargs,
    )

    if _return_in_memory_array:
        return tuple(a._read_stored() for a in arrays)


def visualize(*arrays, filename="cubed", format=None, optimize_graph=True):
    """Produce a visualization of the computation graph for multiple arrays.

    Parameters
    ----------
    arrays : cubed.CoreArray
        The arrays to include in the visualization.
    filename : str
        The name of the file to write to disk. If the provided ``filename``
        doesn't include an extension, '.svg' will be used by default.
    format : {'png', 'pdf', 'dot', 'svg', 'jpeg', 'jpg'}, optional
        Format in which to write output file.  Default is 'svg'.
    optimize_graph : bool, optional
        If True, the graph is optimized before rendering.  Otherwise,
        the graph is displayed as is. Default is True.

    Returns
    -------
    IPython.display.SVG, or None
        An IPython SVG image if IPython can be imported (for rendering
        in a notebook), otherwise None.
    """
    plan = arrays_to_plan(*arrays)
    return plan.visualize(
        filename=filename, format=format, optimize_graph=optimize_graph
    )


class PeakMemoryCallback(Callback):
    def on_task_end(self, event):
        self.peak_memory = event.peak_memory_end


def measure_reserved_memory(executor):
    """Measures the reserved memory use for a given executor runtime.

    This is the memory used by the Python process for running a task,
    excluding any memory used for data for the computation. It can vary by
    operating system, Python version, executor runtime, and installed package
    versions.

    It can be used as a guide to set ``reserved_mem`` when creating a ``Spec`` object.

    This implementation works by running a trivial computation on a tiny amount of
    data and measuring the peak memory use.

    Parameters
    ----------
    executor : cubed.runtime.types.Executor
        The executor to use to run the computation. It must be an executor that
        reports peak memory, such as Lithops or Modal.

    Returns
    -------
    int
        The memory used by the runtime in bytes.
    """
    import cubed.array_api as xp

    a = xp.ones((1,))
    b = xp.negative(a)
    peak_memory_callback = PeakMemoryCallback()
    b.compute(executor=executor, callbacks=[peak_memory_callback])
    return peak_memory_callback.peak_memory



import time

import pytest

pytest.importorskip("lithops")

from lithops.executors import LocalhostExecutor

from cubed.runtime.executors.lithops import map_unordered

# Functions read and write from local filesystem since lithops runs tasks in a separate process


def read_int_from_file(path):
    with open(path) as f:
        return int(f.read())


def write_int_to_file(path, i):
    with open(path, "w") as f:
        f.write(str(i))


def test_map_unordered_no_failures(tmp_path):
    def never_fail(i):
        invocation_count_file = tmp_path / f"{i}"
        write_int_to_file(invocation_count_file, 1)
        return i

    with LocalhostExecutor() as executor:
        for _ in map_unordered(executor, never_fail, range(3), max_failures=0):
            pass

    assert read_int_from_file(tmp_path / "0") == 1
    assert read_int_from_file(tmp_path / "1") == 1
    assert read_int_from_file(tmp_path / "2") == 1


def test_map_unordered(tmp_path):
    def fail_on_first_invocation(i):
        invocation_count_file = tmp_path / f"{i}"
        if invocation_count_file.exists():
            count = read_int_from_file(invocation_count_file)
            write_int_to_file(invocation_count_file, count + 1)
        else:
            write_int_to_file(invocation_count_file, 1)
            raise RuntimeError(f"Fail on {i}")
        return i

    with pytest.raises(RuntimeError):
        with LocalhostExecutor() as executor:
            for _ in map_unordered(
                executor, fail_on_first_invocation, [0, 1, 2], max_failures=2
            ):
                pass

    with LocalhostExecutor() as executor:
        for _ in map_unordered(
            executor, fail_on_first_invocation, [3, 4, 5], max_failures=3
        ):
            pass

    assert read_int_from_file(tmp_path / "3") == 2
    assert read_int_from_file(tmp_path / "4") == 2
    assert read_int_from_file(tmp_path / "5") == 2


def test_map_unordered_execution_timeout(tmp_path):
    def sleep_on_first_invocation(i):
        invocation_count_file = tmp_path / f"{i}"
        if invocation_count_file.exists():
            count = read_int_from_file(invocation_count_file)
            write_int_to_file(invocation_count_file, count + 1)
        else:
            write_int_to_file(invocation_count_file, 1)
            # only sleep on first invocation of input = 0
            if i == 0:
                time.sleep(60)
        return i

    # set execution timeout to less than sleep value above, to check that
    # task is retried after timeout exception
    config = {"lithops": {"execution_timeout": 30}}
    with LocalhostExecutor(config=config) as executor:
        for _ in map_unordered(executor, sleep_on_first_invocation, [0, 1, 2]):
            pass

    assert read_int_from_file(tmp_path / "0") == 2
    assert read_int_from_file(tmp_path / "1") == 1
    assert read_int_from_file(tmp_path / "2") == 1


def test_map_unordered_stragglers(tmp_path):
    def sleep_on_first_invocation(i):
        invocation_count_file = tmp_path / f"{i}"
        if invocation_count_file.exists():
            count = read_int_from_file(invocation_count_file)
            write_int_to_file(invocation_count_file, count + 1)
        else:
            write_int_to_file(invocation_count_file, 1)
            # only sleep on first invocation of input = 0
            if i == 0:
                time.sleep(30)
        return i

    # TODO: run a test like this using a cloud executor
    # Reason: this test passes, but lithops local mode only runs one job at a time,
    # so it actually waits for the first job to finish before running the second one.

    config = {"lithops": {"log_level": "DEBUG"}}
    with LocalhostExecutor(config=config) as executor:
        for _ in map_unordered(
            executor, sleep_on_first_invocation, range(10), use_backups=True
        ):
            pass

    assert read_int_from_file(tmp_path / "0") == 2
    for i in range(1, 10):
        assert read_int_from_file(tmp_path / f"{i}") == 1



from typing import Union

from rechunker.types import PipelineExecutor


class DagExecutor:
    def execute_dag(self, dag, **kwargs):
        raise NotImplementedError  # pragma: no cover


Executor = Union[PipelineExecutor, DagExecutor]



import operator

import numpy as np

from cubed.array_api.creation_functions import asarray
from cubed.array_api.data_type_functions import result_type
from cubed.array_api.dtypes import (
    _boolean_dtypes,
    _dtype_categories,
    _floating_dtypes,
    _integer_or_boolean_dtypes,
    _numeric_dtypes,
)
from cubed.array_api.linear_algebra_functions import matmul
from cubed.core.array import CoreArray
from cubed.core.ops import elemwise


class Array(CoreArray):
    def __init__(self, name, zarray, spec, plan):
        super().__init__(name, zarray, spec, plan)

    def __array__(self, dtype=None) -> np.ndarray:
        x = self.compute()
        if dtype and x.dtype != dtype:
            x = x.astype(dtype)
        if not isinstance(x, np.ndarray):
            x = np.array(x)
        return x

    def __repr__(self):
        return f"cubed.Array<{self.name}, shape={self.shape}, dtype={self.dtype}, chunks={self.chunks}>"

    # Attributes

    @property
    def device(self):
        return "cpu"

    @property
    def mT(self):
        from cubed.array_api.linear_algebra_functions import matrix_transpose

        return matrix_transpose(self)

    @property
    def T(self):
        if self.ndim != 2:
            raise ValueError("x.T requires x to have 2 dimensions.")
        from cubed.array_api.linear_algebra_functions import matrix_transpose

        return matrix_transpose(self)

    # Arithmetic Operators

    def __neg__(self, /):
        if self.dtype not in _numeric_dtypes:
            raise TypeError("Only numeric dtypes are allowed in __neg__")
        return elemwise(np.negative, self, dtype=self.dtype)

    def __pos__(self, /):
        if self.dtype not in _numeric_dtypes:
            raise TypeError("Only numeric dtypes are allowed in __pos__")
        return elemwise(np.positive, self, dtype=self.dtype)

    def __add__(self, other, /):
        other = self._check_allowed_dtypes(other, "numeric", "__add__")
        if other is NotImplemented:
            return other
        return elemwise(np.add, self, other, dtype=result_type(self, other))

    def __sub__(self, other, /):
        other = self._check_allowed_dtypes(other, "numeric", "__sub__")
        if other is NotImplemented:
            return other
        return elemwise(np.subtract, self, other, dtype=result_type(self, other))

    def __mul__(self, other, /):
        other = self._check_allowed_dtypes(other, "numeric", "__mul__")
        if other is NotImplemented:
            return other
        return elemwise(np.multiply, self, other, dtype=result_type(self, other))

    def __truediv__(self, other, /):
        other = self._check_allowed_dtypes(other, "floating-point", "__truediv__")
        if other is NotImplemented:
            return other
        return elemwise(np.divide, self, other, dtype=result_type(self, other))

    def __floordiv__(self, other, /):
        other = self._check_allowed_dtypes(other, "numeric", "__floordiv__")
        if other is NotImplemented:
            return other
        return elemwise(np.floor_divide, self, other, dtype=result_type(self, other))

    def __mod__(self, other, /):
        other = self._check_allowed_dtypes(other, "numeric", "__mod__")
        if other is NotImplemented:
            return other
        return elemwise(np.remainder, self, other, dtype=result_type(self, other))

    def __pow__(self, other, /):
        other = self._check_allowed_dtypes(other, "numeric", "__pow__")
        if other is NotImplemented:
            return other
        return elemwise(np.power, self, other, dtype=result_type(self, other))

    # Array Operators

    def __matmul__(self, other, /):
        other = self._check_allowed_dtypes(other, "numeric", "__matmul__")
        if other is NotImplemented:
            return other
        return matmul(self, other)

    # Bitwise Operators

    def __invert__(self, /):
        if self.dtype not in _integer_or_boolean_dtypes:
            raise TypeError("Only integer or boolean dtypes are allowed in __invert__")
        return elemwise(np.invert, self, dtype=self.dtype)

    def __and__(self, other, /):
        other = self._check_allowed_dtypes(other, "integer or boolean", "__and__")
        if other is NotImplemented:
            return other
        return elemwise(np.bitwise_and, self, other, dtype=result_type(self, other))

    def __or__(self, other, /):
        other = self._check_allowed_dtypes(other, "integer or boolean", "__or__")
        if other is NotImplemented:
            return other
        return elemwise(np.bitwise_or, self, other, dtype=result_type(self, other))

    def __xor__(self, other, /):
        other = self._check_allowed_dtypes(other, "integer or boolean", "__xor__")
        if other is NotImplemented:
            return other
        return elemwise(np.bitwise_xor, self, other, dtype=result_type(self, other))

    def __lshift__(self, other, /):
        other = self._check_allowed_dtypes(other, "integer", "__lshift__")
        if other is NotImplemented:
            return other
        return elemwise(np.left_shift, self, other, dtype=result_type(self, other))

    def __rshift__(self, other, /):
        other = self._check_allowed_dtypes(other, "integer", "__rshift__")
        if other is NotImplemented:
            return other
        return elemwise(np.right_shift, self, other, dtype=result_type(self, other))

    # Comparison Operators

    def __eq__(self, other, /):
        other = self._check_allowed_dtypes(other, "all", "__eq__")
        if other is NotImplemented:
            return other
        return elemwise(np.equal, self, other, dtype=np.bool_)

    def __ge__(self, other, /):
        other = self._check_allowed_dtypes(other, "all", "__ge__")
        if other is NotImplemented:
            return other
        return elemwise(np.greater_equal, self, other, dtype=np.bool_)

    def __gt__(self, other, /):
        other = self._check_allowed_dtypes(other, "all", "__gt__")
        if other is NotImplemented:
            return other
        return elemwise(np.greater, self, other, dtype=np.bool_)

    def __le__(self, other, /):
        other = self._check_allowed_dtypes(other, "all", "__le__")
        if other is NotImplemented:
            return other
        return elemwise(np.less_equal, self, other, dtype=np.bool_)

    def __lt__(self, other, /):
        other = self._check_allowed_dtypes(other, "all", "__lt__")
        if other is NotImplemented:
            return other
        return elemwise(np.less, self, other, dtype=np.bool_)

    def __ne__(self, other, /):
        other = self._check_allowed_dtypes(other, "all", "__ne__")
        if other is NotImplemented:
            return other
        return elemwise(np.not_equal, self, other, dtype=np.bool_)

    # Reflected Operators

    # (Reflected) Arithmetic Operators

    def __radd__(self, other, /):
        other = self._check_allowed_dtypes(other, "numeric", "__radd__")
        if other is NotImplemented:
            return other
        return elemwise(np.add, other, self, dtype=result_type(self, other))

    def __rsub__(self, other, /):
        other = self._check_allowed_dtypes(other, "numeric", "__rsub__")
        if other is NotImplemented:
            return other
        return elemwise(np.subtract, other, self, dtype=result_type(self, other))

    def __rmul__(self, other, /):
        other = self._check_allowed_dtypes(other, "numeric", "__rmul__")
        if other is NotImplemented:
            return other
        return elemwise(np.multiply, other, self, dtype=result_type(self, other))

    def __rtruediv__(self, other, /):
        other = self._check_allowed_dtypes(other, "floating-point", "__rtruediv__")
        if other is NotImplemented:
            return other
        return elemwise(np.divide, other, self, dtype=result_type(self, other))

    def __rfloordiv__(self, other, /):
        other = self._check_allowed_dtypes(other, "numeric", "__rfloordiv__")
        if other is NotImplemented:
            return other
        return elemwise(np.floor_divide, other, self, dtype=result_type(self, other))

    def __rmod__(self, other, /):
        other = self._check_allowed_dtypes(other, "numeric", "__rmod__")
        if other is NotImplemented:
            return other
        return elemwise(np.remainder, other, self, dtype=result_type(self, other))

    def __rpow__(self, other, /):
        other = self._check_allowed_dtypes(other, "numeric", "__rpow__")
        if other is NotImplemented:
            return other
        return elemwise(np.power, other, self, dtype=result_type(self, other))

    # (Reflected) Array Operators

    def __rmatmul__(self, other, /):
        other = self._check_allowed_dtypes(other, "numeric", "__rmatmul__")
        if other is NotImplemented:
            return other
        return matmul(other, self)

    # (Reflected) Bitwise Operators

    def __rand__(self, other, /):
        other = self._check_allowed_dtypes(other, "integer or boolean", "__rand__")
        if other is NotImplemented:
            return other
        return elemwise(np.bitwise_and, other, self, dtype=result_type(self, other))

    def __ror__(self, other, /):
        other = self._check_allowed_dtypes(other, "integer or boolean", "__ror__")
        if other is NotImplemented:
            return other
        return elemwise(np.bitwise_or, other, self, dtype=result_type(self, other))

    def __rxor__(self, other, /):
        other = self._check_allowed_dtypes(other, "integer or boolean", "__rxor__")
        if other is NotImplemented:
            return other
        return elemwise(np.bitwise_xor, other, self, dtype=result_type(self, other))

    def __rlshift__(self, other, /):
        other = self._check_allowed_dtypes(other, "integer", "__rlshift__")
        if other is NotImplemented:
            return other
        return elemwise(np.left_shift, other, self, dtype=result_type(self, other))

    def __rrshift__(self, other, /):
        other = self._check_allowed_dtypes(other, "integer", "__rrshift__")
        if other is NotImplemented:
            return other
        return elemwise(np.right_shift, other, self, dtype=result_type(self, other))

    # Methods

    def __abs__(self, /):
        if self.dtype not in _numeric_dtypes:
            raise TypeError("Only numeric dtypes are allowed in __abs__")
        return elemwise(np.abs, self, dtype=self.dtype)

    def __array_namespace__(self, /, *, api_version=None):
        if api_version is not None and not api_version.startswith("2021."):
            raise ValueError(f"Unrecognized array API version: {api_version!r}")
        import cubed.array_api as array_api

        return array_api

    def __bool__(self, /):
        if self.ndim != 0:
            raise TypeError("bool is only allowed on arrays with 0 dimensions")
        return bool(self.compute())

    def __float__(self, /):
        if self.ndim != 0:
            raise TypeError("float is only allowed on arrays with 0 dimensions")
        return float(self.compute())

    def __index__(self, /):
        if self.ndim != 0:
            raise TypeError("index is only allowed on arrays with 0 dimensions")
        return operator.index(self.compute())

    def __int__(self, /):
        if self.ndim != 0:
            raise TypeError("int is only allowed on arrays with 0 dimensions")
        return int(self.compute())

    # Utility methods

    def _check_allowed_dtypes(self, other, dtype_category, op):
        if self.dtype not in _dtype_categories[dtype_category]:
            raise TypeError(f"Only {dtype_category} dtypes are allowed in {op}")
        if isinstance(other, (int, float, bool)):
            other = self._promote_scalar(other)
        elif isinstance(other, CoreArray):
            if other.dtype not in _dtype_categories[dtype_category]:
                raise TypeError(f"Only {dtype_category} dtypes are allowed in {op}")
        else:
            return NotImplemented

        # TODO: more from numpy.array_api

        return other

    def _promote_scalar(self, scalar):
        if isinstance(scalar, bool):
            if self.dtype not in _boolean_dtypes:
                raise TypeError(
                    "Python bool scalars can only be promoted with bool arrays"
                )
        elif isinstance(scalar, int):
            if self.dtype in _boolean_dtypes:
                raise TypeError(
                    "Python int scalars cannot be promoted with bool arrays"
                )
        elif isinstance(scalar, float):
            if self.dtype not in _floating_dtypes:
                raise TypeError(
                    "Python float scalars can only be promoted with floating-point arrays."
                )
        else:
            raise TypeError("'scalar' must be a Python scalar")
        return asarray(scalar, dtype=self.dtype, spec=self.spec)



import numpy as np
from dask.array.reductions import numel

from cubed.array_api.dtypes import (
    _numeric_dtypes,
    _signed_integer_dtypes,
    _unsigned_integer_dtypes,
)
from cubed.core import reduction


def max(x, /, *, axis=None, keepdims=False):
    if x.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in max")
    return reduction(x, np.max, axis=axis, dtype=x.dtype, keepdims=keepdims)


def mean(x, /, *, axis=None, keepdims=False):
    # This implementation uses NumPy and Zarr's structured arrays to store a
    # pair of fields needed to keep per-chunk counts and totals for computing
    # the mean. Structured arrays are row-based, so are less efficient than
    # regular arrays, but for a function that reduces the amount of data stored,
    # this is usually OK. An alternative would be to add support for multiple
    # outputs.
    dtype = x.dtype
    intermediate_dtype = [("n", np.int64), ("total", np.float64)]
    return reduction(
        x,
        _mean_func,
        combine_func=_mean_combine,
        aggegrate_func=_mean_aggregate,
        axis=axis,
        intermediate_dtype=intermediate_dtype,
        dtype=dtype,
        keepdims=keepdims,
    )


def _mean_func(a, **kwargs):
    n = numel(a, **kwargs)
    total = np.sum(a, **kwargs)
    return {"n": n, "total": total}


def _mean_combine(a, **kwargs):
    n = np.sum(a["n"], **kwargs)
    total = np.sum(a["total"], **kwargs)
    return {"n": n, "total": total}


def _mean_aggregate(a):
    return np.divide(a["total"], a["n"])


def min(x, /, *, axis=None, keepdims=False):
    if x.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in min")
    return reduction(x, np.min, axis=axis, dtype=x.dtype, keepdims=keepdims)


def prod(x, /, *, axis=None, dtype=None, keepdims=False):
    if x.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in prod")
    if dtype is None:
        if x.dtype in _signed_integer_dtypes:
            dtype = np.int64
        elif x.dtype in _unsigned_integer_dtypes:
            dtype = np.uint64
        elif x.dtype == np.float32:
            dtype = np.float64
        else:
            dtype = x.dtype
    return reduction(x, np.prod, axis=axis, dtype=dtype, keepdims=keepdims)


def sum(x, /, *, axis=None, dtype=None, keepdims=False):
    if x.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in sum")
    if dtype is None:
        if x.dtype in _signed_integer_dtypes:
            dtype = np.int64
        elif x.dtype in _unsigned_integer_dtypes:
            dtype = np.uint64
        elif x.dtype == np.float32:
            dtype = np.float64
        else:
            dtype = x.dtype
    return reduction(x, np.sum, axis=axis, dtype=dtype, keepdims=keepdims)



import math


def should_launch_backup(
    task,
    now,
    start_times,
    end_times,
    min_tasks=10,
    min_completed_fraction=0.5,
    slow_factor=3.0,
):
    """
    Determine whether to launch a backup task.

    Backup tasks are only launched if there are at least `min_tasks` being run, and `min_completed_fraction` of tasks have completed.
    If both those criteria have been met, then a backup task is launched if the duration of the current task is at least
    `slow_factor` times slower than the `min_completed_fraction` percentile task duration.
    """
    if len(start_times) < min_tasks:
        return False
    n = math.ceil(len(start_times) * min_completed_fraction) - 1
    if len(end_times) <= n:
        return False
    completed_durations = sorted(
        [end_times[task] - start_times[task] for task in end_times]
    )
    duration = now - start_times[task]
    return duration > completed_durations[n] * slow_factor



import numpy as np

from cubed.array_api.data_type_functions import result_type
from cubed.array_api.dtypes import _numeric_dtypes
from cubed.array_api.manipulation_functions import reshape
from cubed.core.ops import arg_reduction, elemwise


def argmax(x, /, *, axis=None, keepdims=False):
    if x.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in argmax")
    if axis is None:
        x = reshape(x, (-1,))
        axis = 0
        keepdims = False
    return arg_reduction(x, np.argmax, axis=axis, keepdims=keepdims)


def argmin(x, /, *, axis=None, keepdims=False):
    if x.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in argmin")
    if axis is None:
        x = reshape(x, (-1,))
        axis = 0
        keepdims = False
    return arg_reduction(x, np.argmin, axis=axis, keepdims=keepdims)


def where(condition, x1, x2, /):
    dtype = result_type(x1, x2)
    return elemwise(np.where, condition, x1, x2, dtype=dtype)



import pytest

modal = pytest.importorskip("modal")

import asyncio

import fsspec
import modal.aio

from cubed.runtime.executors.modal_async import map_unordered
from cubed.utils import join_path


def read_int_from_file(path):
    with fsspec.open(path) as f:
        return int(f.read())


def write_int_to_file(path, i):
    with fsspec.open(path, "w") as f:
        f.write(str(i))


tmp_path = "s3://cubed-unittest/map_unordered"


stub = modal.aio.AioStub()

image = modal.DebianSlim().pip_install(
    [
        "dask[array]",
        "fsspec",
        "networkx",
        "pytest-mock",  # TODO: only needed for tests
        "rechunker",
        "s3fs",
        "tenacity",
        "zarr",
    ]
)


@stub.function(image=image, secret=modal.ref("my-aws-secret"))
def never_fail(i):
    invocation_count_file = join_path(tmp_path, f"{i}")
    write_int_to_file(invocation_count_file, 1)
    return i


@stub.function(image=image, secret=modal.ref("my-aws-secret"), retries=2)
async def fail_on_first_invocation(i):
    invocation_count_file = join_path(tmp_path, f"{i}")
    fs = fsspec.open(invocation_count_file).fs
    if fs.exists(invocation_count_file):
        count = read_int_from_file(invocation_count_file)
        write_int_to_file(invocation_count_file, count + 1)
    else:
        write_int_to_file(invocation_count_file, 1)
        raise RuntimeError(f"Fail on {i}")
    return i


@stub.function(image=image, secret=modal.ref("my-aws-secret"))
async def fail_on_first_invocation_no_retry(i):
    invocation_count_file = join_path(tmp_path, f"{i}")
    fs = fsspec.open(invocation_count_file).fs
    if fs.exists(invocation_count_file):
        count = read_int_from_file(invocation_count_file)
        write_int_to_file(invocation_count_file, count + 1)
    else:
        write_int_to_file(invocation_count_file, 1)
        raise RuntimeError(f"Fail on {i}")
    return i


@stub.function(image=image, secret=modal.ref("my-aws-secret"))
async def sleep_on_first_invocation(i):
    print(f"sleep_on_first_invocation {i}")
    invocation_count_file = join_path(tmp_path, f"{i}")
    fs = fsspec.open(invocation_count_file).fs
    if fs.exists(invocation_count_file):
        count = read_int_from_file(invocation_count_file)
        write_int_to_file(invocation_count_file, count + 1)
    else:
        write_int_to_file(invocation_count_file, 1)
        # only sleep on first invocation of input = 0
        if i == 0:
            print("sleeping...")
            await asyncio.sleep(60)
            print("... finished sleeping")
    return i


async def run_test(app_function, input, max_failures=3, use_backups=False):
    async with stub.run():
        async for _ in map_unordered(
            app_function, input, max_failures=max_failures, use_backups=use_backups
        ):
            pass


@pytest.mark.cloud
def test_map_unordered_no_failures():
    try:
        asyncio.run(run_test(app_function=never_fail, input=range(3)))

        assert read_int_from_file(join_path(tmp_path, "0")) == 1
        assert read_int_from_file(join_path(tmp_path, "1")) == 1
        assert read_int_from_file(join_path(tmp_path, "2")) == 1

    finally:
        fs = fsspec.open(tmp_path).fs
        fs.rm(tmp_path, recursive=True)


@pytest.mark.cloud
def test_map_unordered_recovers_from_failures():
    try:
        asyncio.run(run_test(app_function=fail_on_first_invocation, input=range(3)))

        assert read_int_from_file(join_path(tmp_path, "0")) == 2
        assert read_int_from_file(join_path(tmp_path, "1")) == 2
        assert read_int_from_file(join_path(tmp_path, "2")) == 2

    finally:
        fs = fsspec.open(tmp_path).fs
        fs.rm(tmp_path, recursive=True)


@pytest.mark.cloud
def test_map_unordered_too_many_failures():
    try:
        with pytest.raises(RuntimeError):
            asyncio.run(
                run_test(
                    app_function=fail_on_first_invocation_no_retry,
                    input=range(3),
                )
            )

    finally:
        fs = fsspec.open(tmp_path).fs
        fs.rm(tmp_path, recursive=True)


@pytest.mark.cloud
def test_map_unordered_stragglers():
    try:
        asyncio.run(
            run_test(
                app_function=sleep_on_first_invocation,
                input=range(10),
                use_backups=True,
            )
        )

        assert read_int_from_file(join_path(tmp_path, "0")) == 2
        for i in range(1, 10):
            assert read_int_from_file(join_path(tmp_path, f"{i}")) == 1

    finally:
        fs = fsspec.open(tmp_path).fs
        fs.rm(tmp_path, recursive=True)



import itertools
from dataclasses import dataclass
from functools import partial
from typing import Callable, Dict

import toolz
import zarr
from dask.array.core import normalize_chunks
from dask.blockwise import _get_coord_mapping, _make_dims, lol_product
from dask.core import flatten
from rechunker.types import ArrayProxy, Stage
from toolz import map

from cubed.utils import chunk_memory, get_item, to_chunksize

from .types import CubedPipeline

sym_counter = 0


def gensym(name):
    global sym_counter
    sym_counter += 1
    return f"{name}-{sym_counter:03}"


@dataclass(frozen=True)
class BlockwiseSpec:
    """Specification for how to run blockwise on an array.

    This is similar to ``CopySpec`` in rechunker.

    Attributes
    ----------
    block_function : Callable
        A function that maps input chunk indexes to an output chunk index.
    function : Callable
        A function that maps input chunks to an output chunk.
    reads_map : Dict[str, ArrayProxy]
        Read proxy dictionary keyed by array name.
    write : ArrayProxy
        Write proxy with an ``array`` attribute that supports ``__setitem__``.
    """

    block_function: Callable
    function: Callable
    reads_map: Dict[str, ArrayProxy]
    write: ArrayProxy


def apply_blockwise(out_key, *, config=BlockwiseSpec):
    """Stage function for blockwise."""
    # lithops needs params to be lists not tuples, so convert back
    out_key = tuple(out_key)
    out_chunk_key = key_to_slices(out_key, config.write.array, config.write.chunks)
    args = []
    name_chunk_inds = config.block_function(("out",) + out_key)
    for name_chunk_ind in name_chunk_inds:
        name = name_chunk_ind[0]
        chunk_ind = name_chunk_ind[1:]
        arr = config.reads_map[name].array
        chunk_key = key_to_slices(chunk_ind, arr)
        arg = arr[chunk_key]
        args.append(arg)

    result = config.function(*args)
    if isinstance(result, dict):  # structured array with named fields
        for k, v in result.items():
            config.write.array.set_basic_selection(out_chunk_key, v, fields=k)
    else:
        config.write.array[out_chunk_key] = result


def key_to_slices(key, arr, chunks=None):
    """Convert a chunk index key to a tuple of slices"""
    chunks = normalize_chunks(chunks or arr.chunks, shape=arr.shape, dtype=arr.dtype)
    return get_item(chunks, key)


def blockwise(
    func,
    out_ind,
    *args,
    max_mem,
    target_store,
    shape,
    dtype,
    chunks,
    new_axes=None,
    in_names=None,
    out_name=None,
    **kwargs,
):
    """Apply a function across blocks from multiple source Zarr arrays.

    Parameters
    ----------
    func : callable
        Function to apply to individual tuples of blocks
    out_ind : iterable
        Block pattern of the output, something like 'ijk' or (1, 2, 3)
    *args : sequence of Array, index pairs
        Sequence like (x, 'ij', y, 'jk', z, 'i')
    max_mem : int
        Maximum memory allowed for a single task of this operation, measured in bytes
    target_store : string or zarr.Array
        Path to output Zarr store, or Zarr array
    shape : tuple
        The shape of the output array.
    dtype : np.dtype
        The ``dtype`` of the output array.
    chunks : tuple
        The chunks of the output array.
    new_axes : dict
        New indexes and their dimension lengths
    **kwargs : dict
        Extra keyword arguments to pass to function

    Returns
    -------
    pipeline:  Pipeline to run the operation
    target:  ArrayProxy for the Zarr array output
    required_mem: minimum memory required per-task, in bytes
    """

    # Use dask's make_blockwise_graph
    arrays = args[::2]
    array_names = in_names or [f"in_{i}" for i in range(len(arrays))]
    array_map = {name: array for name, array in zip(array_names, arrays)}

    inds = args[1::2]

    numblocks = {}
    for name, array in zip(array_names, arrays):
        input_chunks = normalize_chunks(
            array.chunks, shape=array.shape, dtype=array.dtype
        )
        numblocks[name] = tuple(map(len, input_chunks))

    argindsstr = []
    for name, ind in zip(array_names, inds):
        argindsstr.extend((name, ind))

    # TODO: check output shape and chunks are consistent with inputs
    chunks = normalize_chunks(chunks, shape=shape, dtype=dtype)

    # block func

    block_function = make_blockwise_function_flattened(
        func,
        out_name or "out",
        out_ind,
        *argindsstr,
        numblocks=numblocks,
        new_axes=new_axes,
    )

    output_blocks = get_output_blocks(
        func,
        out_name or "out",
        out_ind,
        *argindsstr,
        numblocks=numblocks,
        new_axes=new_axes,
    )

    # end block func

    chunksize = to_chunksize(chunks)
    if isinstance(target_store, zarr.Array):
        target_array = target_store
    else:
        target_array = zarr.empty(
            shape, store=target_store, chunks=chunksize, dtype=dtype
        )

    func_with_kwargs = partial(func, **kwargs)
    read_proxies = {
        name: ArrayProxy(array, array.chunks) for name, array in array_map.items()
    }
    write_proxy = ArrayProxy(target_array, chunksize)
    spec = BlockwiseSpec(block_function, func_with_kwargs, read_proxies, write_proxy)

    stages = [
        Stage(
            apply_blockwise,
            gensym("apply_blockwise"),
            mappable=output_blocks,
        )
    ]

    # calculate memory requirement
    required_mem = 0
    # inputs
    for array in arrays:  # inputs
        # memory for a compressed and an uncompressed input array chunk
        # - we assume compression has no effect (so it's an overestimate)
        # - ideally we'd be able to look at nbytes_stored,
        #   but this is not possible in general since the array has not been written yet
        required_mem += chunk_memory(array.dtype, array.chunks) * 2
    # output
    # memory for a compressed and an uncompressed output array chunk
    # - this assumes the blockwise function creates a new array)
    # - numcodecs uses a working output buffer that's the size of the array being compressed
    required_mem += chunk_memory(dtype, chunksize) * 2

    if required_mem > max_mem:
        raise ValueError(
            f"Blockwise memory ({required_mem}) exceeds max_mem ({max_mem})"
        )

    num_tasks = len(output_blocks)

    return CubedPipeline(stages, spec, target_array, required_mem, num_tasks)


# Code for fusing pipelines


def is_fuse_candidate(pipeline):
    """
    Return True if a pipeline is a candidate for blockwise fusion.
    """
    stages = pipeline.stages
    return len(stages) == 1 and stages[0].function == apply_blockwise


def can_fuse_pipelines(pipeline1, pipeline2):
    if is_fuse_candidate(pipeline1) and is_fuse_candidate(pipeline2):
        return pipeline1.num_tasks == pipeline2.num_tasks
    return False


def fuse(pipeline1, pipeline2):
    """
    Fuse two blockwise pipelines into a single pipeline, avoiding writing to (or reading from) the target of the first pipeline.
    """

    assert pipeline1.num_tasks == pipeline2.num_tasks

    mappable = pipeline2.stages[0].mappable

    stages = [
        Stage(
            apply_blockwise,
            gensym("fused_apply_blockwise"),
            mappable=mappable,
        )
    ]

    def fused_blockwise_func(out_key):
        return pipeline1.config.block_function(
            *pipeline2.config.block_function(out_key)
        )

    def fused_func(*args):
        return pipeline2.config.function(pipeline1.config.function(*args))

    read_proxies = pipeline1.config.reads_map
    write_proxy = pipeline2.config.write
    spec = BlockwiseSpec(fused_blockwise_func, fused_func, read_proxies, write_proxy)

    target_array = pipeline2.target_array
    required_mem = max(pipeline1.required_mem, pipeline2.required_mem)
    num_tasks = pipeline2.num_tasks

    return CubedPipeline(stages, spec, target_array, required_mem, num_tasks)


# blockwise functions


def make_blockwise_function(
    func, output, out_indices, *arrind_pairs, numblocks=None, new_axes=None
):
    """Make a function that is the equivalent of make_blockwise_graph."""

    if numblocks is None:
        raise ValueError("Missing required numblocks argument.")
    new_axes = new_axes or {}
    argpairs = list(toolz.partition(2, arrind_pairs))

    # Dictionary mapping {i: 3, j: 4, ...} for i, j, ... the dimensions
    dims = _make_dims(argpairs, numblocks, new_axes)

    # Generate the abstract "plan" before constructing
    # the actual graph
    (coord_maps, concat_axes, dummies) = _get_coord_mapping(
        dims,
        output,
        out_indices,
        numblocks,
        argpairs,
        False,
    )

    def blockwise_fn(out_key):
        out_coords = out_key[1:]

        # from Dask make_blockwise_graph
        deps = set()
        coords = out_coords + dummies
        args = []
        for cmap, axes, (arg, ind) in zip(coord_maps, concat_axes, argpairs):
            if ind is None:
                args.append(arg)
            else:
                arg_coords = tuple(coords[c] for c in cmap)
                if axes:
                    tups = lol_product((arg,), arg_coords)
                    deps.update(flatten(tups))
                else:
                    tups = (arg,) + arg_coords
                    deps.add(tups)
                args.append(tups)

        args.insert(0, func)
        val = tuple(args)
        # end from make_blockwise_graph

        return val

    return blockwise_fn


def make_blockwise_function_flattened(
    func, output, out_indices, *arrind_pairs, numblocks=None, new_axes=None
):
    # TODO: make this a part of make_blockwise_function?
    blockwise_fn = make_blockwise_function(
        func, output, out_indices, *arrind_pairs, numblocks=numblocks, new_axes=new_axes
    )

    def blockwise_fn_flattened(out_key):
        name_chunk_inds = blockwise_fn(out_key)[1:]  # drop function in position 0
        # flatten (nested) lists indicating contraction
        if isinstance(name_chunk_inds[0], list):
            name_chunk_inds = list(flatten(name_chunk_inds))
        return name_chunk_inds

    return blockwise_fn_flattened


def get_output_blocks(
    func, output, out_indices, *arrind_pairs, numblocks=None, new_axes=None
):

    if numblocks is None:
        raise ValueError("Missing required numblocks argument.")
    new_axes = new_axes or {}
    argpairs = list(toolz.partition(2, arrind_pairs))

    # Dictionary mapping {i: 3, j: 4, ...} for i, j, ... the dimensions
    dims = _make_dims(argpairs, numblocks, new_axes)

    # return a list of lists, not of tuples, otherwise lithops breaks
    output_blocks_lists = [
        list(tup) for tup in itertools.product(*[range(dims[i]) for i in out_indices])
    ]
    return output_blocks_lists



from tenacity import retry, stop_after_attempt

from cubed.core.array import TaskEndEvent
from cubed.core.plan import visit_nodes
from cubed.runtime.types import DagExecutor


@retry(stop=stop_after_attempt(3))
def exec_stage_func(func, *args, **kwargs):
    return func(*args, **kwargs)


class PythonDagExecutor(DagExecutor):
    """The default execution engine that runs tasks sequentially uses Python loops."""

    def execute_dag(self, dag, callbacks=None, array_names=None, **kwargs):
        for name, node in visit_nodes(dag):
            pipeline = node["pipeline"]
            for stage in pipeline.stages:
                if stage.mappable is not None:
                    for m in stage.mappable:
                        exec_stage_func(stage.function, m, config=pipeline.config)
                        if callbacks is not None:
                            event = TaskEndEvent(array_name=name)
                            [callback.on_task_end(event) for callback in callbacks]
                else:
                    exec_stage_func(stage.function, config=pipeline.config)
                    if callbacks is not None:
                        event = TaskEndEvent(array_name=name)
                        [callback.on_task_end(event) for callback in callbacks]



import asyncio
import time
from concurrent.futures import ThreadPoolExecutor
from functools import partial

from aiostream import stream
from tenacity import retry, stop_after_attempt

from cubed.core.array import TaskEndEvent
from cubed.core.plan import visit_node_generations
from cubed.runtime.types import DagExecutor
from cubed.utils import peak_memory


def execution_stats(func):
    """Measure timing information and peak memory usage of a function call."""

    def wrapper(*args, **kwargs):
        peak_memory_start = peak_memory()
        function_start_tstamp = time.time()
        result = func(*args, **kwargs)
        function_end_tstamp = time.time()
        peak_memory_end = peak_memory()
        return result, dict(
            function_start_tstamp=function_start_tstamp,
            function_end_tstamp=function_end_tstamp,
            peak_memory_start=peak_memory_start,
            peak_memory_end=peak_memory_end,
        )

    return wrapper


@retry(stop=stop_after_attempt(3))
@execution_stats
def run_func(input, func=None, config=None, name=None):
    print(f"{name}: running on {input}")
    result = func(input, config=config)
    return result


async def map_unordered(
    concurrent_executor,
    function,
    input,
    max_failures=3,
    use_backups=False,
    **kwargs,
):
    print(f"{kwargs['name']}: running map_unordered")
    tasks = {
        asyncio.wrap_future(concurrent_executor.submit(function, i, **kwargs)): i
        for i in input
    }
    pending = set(tasks.keys())

    while pending:
        finished, pending = await asyncio.wait(
            pending, return_when=asyncio.FIRST_COMPLETED, timeout=2
        )
        for task in finished:
            if task.exception():
                raise task.exception()
            yield task.result()


async def produce(*args, **kwargs):
    task_create_tstamp = time.time()
    name = kwargs.get("name", None)
    async for result, stats in map_unordered(*args, **kwargs):
        stats["array_name"] = name
        stats["task_create_tstamp"] = task_create_tstamp
        yield result, stats


def pipeline_to_stream(concurrent_executor, name, pipeline, **kwargs):
    if any([stage for stage in pipeline.stages if stage.mappable is None]):
        raise NotImplementedError("All stages must be mappable in pipelines")
    it = stream.iterate(
        [
            partial(
                produce,
                concurrent_executor,
                run_func,
                list(stage.mappable),
                func=stage.function,
                config=pipeline.config,
                name=name,
                **kwargs,
            )
            for stage in pipeline.stages
            if stage.mappable is not None
        ]
    )
    # concat stages, running only one stage at a time
    return stream.concatmap(it, lambda f: f(), task_limit=1)


async def async_execute_dag(dag, callbacks=None, array_names=None, **kwargs):
    with ThreadPoolExecutor() as concurrent_executor:
        for gen in visit_node_generations(dag):
            # run pipelines in the same topological generation in parallel by merging their streams
            streams = [
                pipeline_to_stream(
                    concurrent_executor, name, node["pipeline"], **kwargs
                )
                for name, node in gen
            ]
            merged_stream = stream.merge(*streams)
            async with merged_stream.stream() as streamer:
                async for _, stats in streamer:
                    handle_callbacks(callbacks, stats)


def handle_callbacks(callbacks, stats):
    if callbacks is not None:
        task_result_tstamp = time.time()
        event = TaskEndEvent(
            task_result_tstamp=task_result_tstamp,
            **stats,
        )
        [callback.on_task_end(event) for callback in callbacks]


class AsyncPythonDagExecutor(DagExecutor):
    """An execution engine that uses Python asyncio."""

    def execute_dag(self, dag, callbacks=None, array_names=None, **kwargs):
        asyncio.run(
            async_execute_dag(
                dag, callbacks=callbacks, array_names=array_names, **kwargs
            )
        )



import itertools
import math
from typing import Any, Iterable, Iterator, List, Tuple

import dask
import numpy as np
from rechunker.types import CopySpec, Stage

from cubed.primitive.types import CubedPipeline

from .utils import gensym


# differs from rechunker's chunk_keys to return a list rather than a tuple, to keep lithops happy
def chunk_keys(
    shape: Tuple[int, ...], chunks: Tuple[int, ...]
) -> Iterator[List[slice]]:
    """Iterator over array indexing keys of the desired chunk sized.

    The union of all keys indexes every element of an array of shape ``shape``
    exactly once. Each array resulting from indexing is of shape ``chunks``,
    except possibly for the last arrays along each dimension (if ``chunks``
    do not even divide ``shape``).
    """
    ranges = [range(math.ceil(s / c)) for s, c in zip(shape, chunks)]
    for indices in itertools.product(*ranges):
        yield [
            slice(c * i, min(c * (i + 1), s)) for i, s, c in zip(indices, shape, chunks)
        ]


# need a ChunkKeys Iterable instead of a generator, to avoid beam pickle error
class ChunkKeys(Iterable[Tuple[slice, ...]]):
    def __init__(self, shape: Tuple[int, ...], chunks: Tuple[int, ...]):
        self.shape = shape
        self.chunks = chunks

    def __iter__(self):
        return chunk_keys(self.shape, self.chunks)


def copy_read_to_write(chunk_key, *, config=CopySpec):
    # workaround limitation of lithops.utils.verify_args
    if isinstance(chunk_key, list):
        chunk_key = tuple(chunk_key)
    with dask.config.set(scheduler="single-threaded"):
        data = np.asarray(config.read.array[chunk_key])
    config.write.array[chunk_key] = data


def copy_read_to_intermediate(chunk_key, *, config=CopySpec):
    # workaround limitation of lithops.utils.verify_args
    if isinstance(chunk_key, list):
        chunk_key = tuple(chunk_key)
    with dask.config.set(scheduler="single-threaded"):
        data = np.asarray(config.read.array[chunk_key])
    config.intermediate.array[chunk_key] = data


def copy_intermediate_to_write(chunk_key, *, config=CopySpec):
    # workaround limitation of lithops.utils.verify_args
    if isinstance(chunk_key, list):
        chunk_key = tuple(chunk_key)
    with dask.config.set(scheduler="single-threaded"):
        data = np.asarray(config.intermediate.array[chunk_key])
    config.write.array[chunk_key] = data


def spec_to_pipeline(
    spec: CopySpec, target_array: Any, max_mem: int, num_tasks: int
) -> CubedPipeline:
    # typing won't work until we start using numpy types
    shape = spec.read.array.shape  # type: ignore
    if spec.intermediate.array is None:
        stages = [
            Stage(
                copy_read_to_write,
                gensym("copy_read_to_write"),
                mappable=ChunkKeys(shape, spec.write.chunks),
            )
        ]
    else:
        stages = [
            Stage(
                copy_read_to_intermediate,
                gensym("copy_read_to_intermediate"),
                mappable=ChunkKeys(shape, spec.intermediate.chunks),
            ),
            Stage(
                copy_intermediate_to_write,
                gensym("copy_intermediate_to_write"),
                mappable=ChunkKeys(shape, spec.write.chunks),
            ),
        ]
    return CubedPipeline(stages, spec, target_array, max_mem, num_tasks)


def already_computed(node_dict):
    """
    Return True if the array for a node doesn't have a pipeline to compute it,
    or it has already been computed (all chunks are present).
    """
    pipeline = node_dict.get("pipeline", None)
    if pipeline is None:
        return True
    target = node_dict.get("target", None)
    if target.ndim > 0 and target.nchunks_initialized == target.nchunks:
        return True
    return False



import pandas as pd
import pytest

pytest.importorskip("lithops")

import cubed
import cubed.array_api as xp
import cubed.random
from cubed.extensions.history import HistoryCallback
from cubed.runtime.executors.lithops import LithopsDagExecutor
from cubed.tests.utils import LITHOPS_LOCAL_CONFIG


@pytest.fixture()
def spec(tmp_path, reserved_mem):
    return cubed.Spec(tmp_path, max_mem=2_000_000_000, reserved_mem=reserved_mem)


@pytest.fixture(scope="module")
def reserved_mem():
    executor = LithopsDagExecutor(config=LITHOPS_LOCAL_CONFIG)
    return cubed.measure_reserved_memory(executor) * 1.05  # add some wiggle room


# Array Object


@pytest.mark.slow
def test_index(spec):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = a[1:, :]
    run_operation("index", b)


# Creation Functions


@pytest.mark.slow
def test_eye(spec):
    a = xp.eye(10000, 10000, chunks=(5000, 5000), spec=spec)
    run_operation("eye", a)


@pytest.mark.slow
def test_tril(spec):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = xp.tril(a)
    run_operation("tril", b)


# Elementwise Functions


@pytest.mark.slow
def test_add(spec):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    c = xp.add(a, b)
    run_operation("add", c)


@pytest.mark.slow
def test_negative(spec):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = xp.negative(a)
    run_operation("negative", b)


# Linear Algebra Functions


@pytest.mark.slow
def test_matmul(spec):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    c = xp.astype(a, xp.float32)
    d = xp.astype(b, xp.float32)
    e = xp.matmul(c, d)
    run_operation("matmul", e)


@pytest.mark.slow
def test_matrix_transpose(spec):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = xp.matrix_transpose(a)
    run_operation("matrix_transpose", b)


@pytest.mark.slow
def test_tensordot(spec):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    c = xp.astype(a, xp.float32)
    d = xp.astype(b, xp.float32)
    e = xp.tensordot(c, d, axes=1)
    run_operation("tensordot", e)


# Manipulation Functions


@pytest.mark.slow
def test_concat(spec):
    # Note 'a' has one fewer element in axis=0 to force chunking to cross array boundaries
    a = cubed.random.random(
        (9999, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    c = xp.concat((a, b), axis=0)
    run_operation("concat", c)


@pytest.mark.slow
def test_reshape(spec):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    # need intermediate reshape due to limitations in Dask's reshape_rechunk
    b = xp.reshape(a, (5000, 2, 10000))
    c = xp.reshape(b, (5000, 20000))
    run_operation("reshape", c)


@pytest.mark.slow
def test_stack(spec):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    c = xp.stack((a, b), axis=0)
    run_operation("stack", c)


# Searching Functions


@pytest.mark.slow
def test_argmax(spec):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = xp.argmax(a, axis=0)
    run_operation("argmax", b)


# Statistical Functions


@pytest.mark.slow
def test_max(spec):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = xp.max(a, axis=0)
    run_operation("max", b)


@pytest.mark.slow
def test_mean(spec):
    a = cubed.random.random(
        (10000, 10000), chunks=(5000, 5000), spec=spec
    )  # 200MB chunks
    b = xp.mean(a, axis=0)
    run_operation("mean", b)


# Internal functions


def run_operation(name, result_array):
    # result_array.visualize(f"cubed-{name}-unoptimized", optimize_graph=False)
    # result_array.visualize(f"cubed-{name}")
    executor = LithopsDagExecutor(config=LITHOPS_LOCAL_CONFIG)
    hist = HistoryCallback()
    # use store=None to write to temporary zarr
    cubed.to_zarr(result_array, store=None, executor=executor, callbacks=[hist])

    plan_df = pd.read_csv(hist.plan_df_path)
    stats_df = pd.read_csv(hist.stats_df_path)
    df = analyze(plan_df, stats_df, result_array.spec.reserved_mem)
    print(df)

    # check utilization does not exceed 1
    assert (df["utilization"] <= 1.0).all()


def analyze(plan_df, stats_df, reserved_mem):

    reserved_mem_mb = reserved_mem / 1_000_000
    reserved_mem_mb *= 1.05  # add some wiggle room

    # convert memory to MB
    plan_df["required_mem_mb"] = plan_df["required_mem"] / 1_000_000
    plan_df["total_mem_mb"] = plan_df["required_mem_mb"] + reserved_mem_mb
    plan_df = plan_df[
        ["array_name", "op_name", "required_mem_mb", "total_mem_mb", "num_tasks"]
    ]
    stats_df["peak_mem_start_mb"] = stats_df["peak_memory_start"] / 1_000_000
    stats_df["peak_mem_end_mb"] = stats_df["peak_memory_end"] / 1_000_000
    stats_df["peak_mem_delta_mb"] = (
        stats_df["peak_mem_end_mb"] - stats_df["peak_mem_start_mb"]
    )

    # find per-array stats
    df = stats_df.groupby("array_name", as_index=False).agg(
        {
            "peak_mem_start_mb": ["min", "mean", "max"],
            "peak_mem_end_mb": ["max"],
            "peak_mem_delta_mb": ["min", "mean", "max"],
        }
    )

    # flatten multi-index
    df.columns = ["_".join(a).rstrip("_") for a in df.columns.to_flat_index()]
    df = df.merge(plan_df, on="array_name")

    def utilization(row):
        return row["peak_mem_end_mb_max"] / row["total_mem_mb"]

    df["utilization"] = df.apply(lambda row: utilization(row), axis=1)
    df = df[
        [
            "array_name",
            "op_name",
            "num_tasks",
            "peak_mem_start_mb_max",
            "peak_mem_end_mb_max",
            "peak_mem_delta_mb_max",
            "required_mem_mb",
            "total_mem_mb",
            "utilization",
        ]
    ]

    return df



import dill
import fsspec
import numpy as np
import pytest
import zarr
from numpy.testing import assert_array_equal

import cubed
import cubed.array_api as xp
from cubed.extensions.tqdm import TqdmProgressBar
from cubed.primitive.blockwise import apply_blockwise
from cubed.runtime.types import DagExecutor
from cubed.tests.utils import MAIN_EXECUTORS, MODAL_EXECUTORS, TaskCounter, create_zarr


@pytest.fixture()
def spec(tmp_path):
    return cubed.Spec(tmp_path, max_mem=100000)


@pytest.fixture(scope="module", params=MAIN_EXECUTORS)
def executor(request):
    return request.param


@pytest.fixture(scope="module", params=MODAL_EXECUTORS)
def modal_executor(request):
    return request.param


def test_regular_chunks(spec):
    xp.ones((5, 5), chunks=((2, 2, 1), (5,)), spec=spec)
    with pytest.raises(ValueError):
        xp.ones((5, 5), chunks=((2, 1, 2), (5,)), spec=spec)


class WrappedArray:
    def __init__(self, x):
        self.x = x
        self.dtype = x.dtype
        self.shape = x.shape
        self.ndim = len(x.shape)

    def __array__(self, dtype=None):
        return np.asarray(self.x, dtype=dtype)

    def __getitem__(self, i):
        return WrappedArray(self.x[i])


@pytest.mark.parametrize(
    "x,chunks,asarray",
    [
        (np.arange(25).reshape((5, 5)), (5, 5), None),
        (np.arange(25).reshape((5, 5)), (3, 2), True),
        (np.arange(25).reshape((5, 5)), -1, True),
        (np.array([[1]]), 1, None),
    ],
)
def test_from_array(x, chunks, asarray):
    a = cubed.from_array(WrappedArray(x), chunks=chunks, asarray=asarray)
    assert isinstance(a, cubed.Array)
    assert_array_equal(a, x)


def test_from_array_zarr(tmp_path, spec):
    store = store = tmp_path / "source.zarr"
    za = create_zarr(
        [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
        chunks=(2, 2),
        store=store,
    )
    a = cubed.from_array(za, spec=spec)
    assert_array_equal(a, za)


def test_from_zarr(tmp_path, spec, executor):
    store = store = tmp_path / "source.zarr"
    create_zarr(
        [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
        chunks=(2, 2),
        store=store,
    )
    a = cubed.from_zarr(store, spec=spec)
    assert_array_equal(
        a.compute(executor=executor), np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    )


def test_store(tmp_path, spec):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)

    store = tmp_path / "source.zarr"
    target = zarr.empty(a.shape, store=store)

    cubed.store(a, target)
    assert_array_equal(target, np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]))


def test_store_multiple(tmp_path, spec):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.asarray([[1, 1, 1], [1, 1, 1], [1, 1, 1]], chunks=(2, 2), spec=spec)

    store1 = tmp_path / "source1.zarr"
    target1 = zarr.empty(a.shape, store=store1)
    store2 = tmp_path / "source2.zarr"
    target2 = zarr.empty(b.shape, store=store2)

    cubed.store([a, b], [target1, target2])
    assert_array_equal(target1, np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]))
    assert_array_equal(target2, np.array([[1, 1, 1], [1, 1, 1], [1, 1, 1]]))


def test_store_fails(tmp_path, spec):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    store = tmp_path / "source.zarr"
    target = zarr.empty(a.shape, store=store)

    with pytest.raises(
        ValueError, match=r"Different number of sources \(2\) and targets \(1\)"
    ):
        cubed.store([a, b], [target])

    with pytest.raises(ValueError, match="All sources must be cubed array objects"):
        cubed.store([1], [target])


def test_to_zarr(tmp_path, spec, executor):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    output = tmp_path / "output.zarr"
    cubed.to_zarr(a, output, executor=executor)
    res = zarr.open(output)
    assert_array_equal(res[:], np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]))


def test_map_blocks_with_kwargs(spec, executor):
    # based on dask test
    a = xp.asarray([0, 1, 2, 3, 4, 5, 6, 7, 8, 9], chunks=5, spec=spec)
    b = cubed.map_blocks(np.max, a, axis=0, keepdims=True, dtype=a.dtype, chunks=(1,))
    assert_array_equal(b.compute(executor=executor), np.array([4, 9]))


def test_map_blocks_with_block_id(spec, executor):
    # based on dask test
    def func(block, block_id=None, c=0):
        return np.ones_like(block) * sum(block_id) + c

    a = xp.arange(10, dtype="int64", chunks=(2,))
    b = cubed.map_blocks(func, a, dtype="int64")

    assert_array_equal(
        b.compute(executor=executor),
        np.array([0, 0, 1, 1, 2, 2, 3, 3, 4, 4], dtype="int64"),
    )

    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = cubed.map_blocks(func, a, dtype="int64")

    assert_array_equal(
        b.compute(executor=executor),
        np.array([[0, 0, 1], [0, 0, 1], [1, 1, 2]], dtype="int64"),
    )

    c = cubed.map_blocks(func, a, dtype="int64", c=1)

    assert_array_equal(
        c.compute(executor=executor),
        np.array([[0, 0, 1], [0, 0, 1], [1, 1, 2]], dtype="int64") + 1,
    )


def test_map_blocks_with_different_block_shapes(spec):
    def func(x, y):
        return x

    a = xp.asarray([[[12, 13]]], spec=spec)
    b = xp.asarray([14, 15], spec=spec)
    c = cubed.map_blocks(
        func, a, b, dtype="int64", chunks=(1, 1, 2), drop_axis=2, new_axis=2
    )
    assert_array_equal(c.compute(), np.array([[[12, 13]]]))


def test_multiple_ops(spec, executor):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.asarray([[1, 1, 1], [1, 1, 1], [1, 1, 1]], chunks=(2, 2), spec=spec)
    c = xp.add(a, b)
    d = xp.negative(c)
    assert_array_equal(
        d.compute(executor=executor),
        np.array([[-2, -3, -4], [-5, -6, -7], [-8, -9, -10]]),
    )


def test_rechunk(spec, executor):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 1), spec=spec)
    b = a.rechunk((1, 2))
    assert_array_equal(
        b.compute(executor=executor),
        np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]),
    )


def test_rechunk_same_chunks(spec):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 1), spec=spec)
    b = a.rechunk((2, 1))
    task_counter = TaskCounter()
    res = b.compute(callbacks=[task_counter])
    # no tasks should have run since chunks are same
    assert task_counter.value == 0

    assert_array_equal(res, np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]))


def test_compute_is_idempotent(spec, executor):
    a = xp.ones((3, 3), chunks=(2, 2), spec=spec)
    b = xp.negative(a)
    assert_array_equal(b.compute(executor=executor), -np.ones((3, 3)))
    assert_array_equal(b.compute(executor=executor), -np.ones((3, 3)))


def test_default_spec(executor):
    # default spec works for small computations
    a = xp.ones((3, 3), chunks=(2, 2))
    b = xp.negative(a)
    assert_array_equal(
        b.compute(executor=executor),
        -np.ones((3, 3)),
    )


def test_default_spec_max_mem_exceeded():
    # default spec fails for large computations
    a = xp.ones((100000, 100000), chunks=(10000, 10000))
    with pytest.raises(ValueError):
        xp.negative(a)


def test_different_specs(tmp_path):
    spec1 = cubed.Spec(tmp_path, max_mem=100000)
    spec2 = cubed.Spec(tmp_path, max_mem=200000)
    a = xp.ones((3, 3), chunks=(2, 2), spec=spec1)
    b = xp.ones((3, 3), chunks=(2, 2), spec=spec2)
    with pytest.raises(ValueError):
        xp.add(a, b)


def test_reduction_multiple_rounds(tmp_path, executor):
    spec = cubed.Spec(tmp_path, max_mem=1000)
    a = xp.ones((100, 10), dtype=np.uint8, chunks=(1, 10), spec=spec)
    b = xp.sum(a, axis=0, dtype=np.uint8)
    # check that there is > 1 rechunk step
    rechunks = [
        n for (n, d) in b.plan.dag.nodes(data=True) if d["op_name"] == "rechunk"
    ]
    assert len(rechunks) > 1
    assert b.plan.max_required_mem() == 1000
    assert_array_equal(b.compute(executor=executor), np.ones((100, 10)).sum(axis=0))


def test_reduction_not_enough_memory(tmp_path):
    spec = cubed.Spec(tmp_path, max_mem=50)
    a = xp.ones((100, 10), dtype=np.uint8, chunks=(1, 10), spec=spec)
    with pytest.raises(ValueError, match=r"Not enough memory for reduction"):
        xp.sum(a, axis=0, dtype=np.uint8)


def test_compute_multiple():
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2))
    b = xp.asarray([[1, 1, 1], [1, 1, 1], [1, 1, 1]], chunks=(2, 2))
    c = xp.add(a, b)
    d = c * 2
    e = c * 3

    f = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2))
    g = f * 4

    dc, ec, gc = cubed.compute(d, e, g)

    an = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    bn = np.array([[1, 1, 1], [1, 1, 1], [1, 1, 1]])
    cn = an + bn
    dn = cn * 2
    en = cn * 3

    fn = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    gn = fn * 4

    assert_array_equal(dc, dn)
    assert_array_equal(ec, en)
    assert_array_equal(gc, gn)


def test_compute_multiple_different_specs(tmp_path):
    spec1 = cubed.Spec(tmp_path, max_mem=100000)
    spec2 = cubed.Spec(tmp_path, max_mem=200000)

    a1 = xp.ones((3, 3), chunks=(2, 2), spec=spec1)
    b1 = xp.ones((3, 3), chunks=(2, 2), spec=spec1)
    c1 = xp.add(a1, b1)

    a2 = xp.ones((3, 3), chunks=(2, 2), spec=spec2)
    b2 = xp.ones((3, 3), chunks=(2, 2), spec=spec2)
    c2 = xp.add(a2, b2)

    with pytest.raises(ValueError):
        cubed.compute(c1, c2)


def test_visualize(tmp_path):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2))
    b = xp.asarray([[1, 1, 1], [1, 1, 1], [1, 1, 1]], chunks=(2, 2))
    c = xp.add(a, b)
    d = c * 2
    e = c * 3

    f = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2))
    g = f * 4

    assert not (tmp_path / "e.dot").exists()
    assert not (tmp_path / "e.png").exists()
    assert not (tmp_path / "e.svg").exists()
    assert not (tmp_path / "dg.svg").exists()

    e.visualize(filename=tmp_path / "e")
    assert (tmp_path / "e.svg").exists()

    e.visualize(filename=tmp_path / "e", format="png")
    assert (tmp_path / "e.png").exists()

    e.visualize(filename=tmp_path / "e", format="dot")
    assert (tmp_path / "e.dot").exists()

    # multiple arrays
    cubed.visualize(d, g, filename=tmp_path / "dg")
    assert (tmp_path / "dg.svg").exists()


def test_array_pickle(spec, executor):
    a = xp.asarray(
        [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]],
        chunks=(2, 2),
        spec=spec,
    )
    b = xp.asarray(
        [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]],
        chunks=(2, 2),
        spec=spec,
    )
    c = xp.matmul(a, b)

    # we haven't computed c yet, so pickle and unpickle, and check it still works
    # note we have to use dill which can serialize local functions, unlike pickle
    c = dill.loads(dill.dumps(c))

    x = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]])
    y = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]])
    expected = np.matmul(x, y)
    assert_array_equal(c.compute(executor=executor), expected)


mock_call_counter = 0


def mock_apply_blockwise(*args, **kwargs):
    # Raise an error on every 3rd call
    global mock_call_counter
    mock_call_counter += 1
    if mock_call_counter % 3 == 0:
        raise IOError("Test fault injection")
    return apply_blockwise(*args, **kwargs)


def test_retries(mocker, spec):
    # Inject faults into the primitive layer
    mocker.patch(
        "cubed.primitive.blockwise.apply_blockwise", side_effect=mock_apply_blockwise
    )

    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.asarray([[1, 1, 1], [1, 1, 1], [1, 1, 1]], chunks=(2, 2), spec=spec)
    c = xp.add(a, b)
    assert_array_equal(c.compute(), np.array([[2, 3, 4], [5, 6, 7], [8, 9, 10]]))


def test_callbacks(spec, executor):
    if not isinstance(executor, DagExecutor):
        pytest.skip(f"{type(executor)} does not support callbacks")

    task_counter = TaskCounter()
    progress = TqdmProgressBar()  # test indirectly (doesn't fail)

    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.asarray([[1, 1, 1], [1, 1, 1], [1, 1, 1]], chunks=(2, 2), spec=spec)
    c = xp.add(a, b)
    assert_array_equal(
        c.compute(executor=executor, callbacks=[task_counter, progress]),
        np.array([[2, 3, 4], [5, 6, 7], [8, 9, 10]]),
    )

    assert task_counter.value == 4


@pytest.mark.cloud
def test_callbacks_modal(spec, modal_executor):
    task_counter = TaskCounter()
    tmp_path = "s3://cubed-unittest/callbacks"
    spec = cubed.Spec(tmp_path, max_mem=100000)
    try:
        a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
        b = xp.asarray([[1, 1, 1], [1, 1, 1], [1, 1, 1]], chunks=(2, 2), spec=spec)
        c = xp.add(a, b)
        assert_array_equal(
            c.compute(executor=modal_executor, callbacks=[task_counter]),
            np.array([[2, 3, 4], [5, 6, 7], [8, 9, 10]]),
        )

        assert task_counter.value == 4
    finally:
        fs = fsspec.open(tmp_path).fs
        fs.rm(tmp_path, recursive=True)


def test_already_computed(spec):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.asarray([[1, 1, 1], [1, 1, 1], [1, 1, 1]], chunks=(2, 2), spec=spec)
    c = xp.add(a, b)
    d = xp.negative(c)

    assert d.plan.num_tasks(optimize_graph=False) == 8

    task_counter = TaskCounter()
    c.compute(callbacks=[task_counter], optimize_graph=False)
    assert task_counter.value == 4

    # since c has already been computed, when computing d only 4 tasks are run, instead of 8
    task_counter = TaskCounter()
    d.compute(callbacks=[task_counter], optimize_graph=False)
    assert task_counter.value == 4


def test_measure_reserved_memory(executor):
    from cubed.runtime.executors.lithops import LithopsDagExecutor

    if not isinstance(executor, LithopsDagExecutor):
        pytest.skip(f"{type(executor)} does not support measure_reserved_memory")

    reserved_memory = cubed.measure_reserved_memory(executor=executor)
    assert reserved_memory > 1_000_000  # over 1MB



import os
import time
from asyncio.exceptions import TimeoutError

import modal
from modal.exception import ConnectionError
from tenacity import retry, retry_if_exception_type, stop_after_attempt

from cubed.core.array import TaskEndEvent
from cubed.core.plan import visit_nodes
from cubed.runtime.types import DagExecutor
from cubed.utils import peak_memory

stub = modal.Stub()

requirements_file = os.getenv("CUBED_MODAL_REQUIREMENTS_FILE")

if requirements_file:
    image = modal.DebianSlim().pip_install_from_requirements(requirements_file)
else:
    image = modal.DebianSlim().pip_install(
        [
            "dask[array]",
            "fsspec",
            "networkx",
            "pytest-mock",  # TODO: only needed for tests
            "rechunker",
            "s3fs",
            "tenacity",
            "zarr",
        ]
    )


# Use a generator, since we want results to be returned as they finish and we don't care about order
@stub.generator(image=image, secret=modal.ref("my-aws-secret"), memory=2000, retries=2)
def run_remotely(input, func=None, config=None):
    print(f"running remotely on {input}")
    peak_memory_start = peak_memory()
    function_start_tstamp = time.time()
    result = func(input, config=config)
    function_end_tstamp = time.time()
    peak_memory_end = peak_memory()
    yield (
        result,
        function_start_tstamp,
        function_end_tstamp,
        peak_memory_start,
        peak_memory_end,
    )


# This just retries the initial connection attempt, not the function calls
@retry(
    retry=retry_if_exception_type((TimeoutError, ConnectionError)),
    stop=stop_after_attempt(3),
)
def execute_dag(dag, callbacks=None, array_names=None, **kwargs):
    with stub.run():
        for name, node in visit_nodes(dag):
            pipeline = node["pipeline"]

            for stage in pipeline.stages:
                if stage.mappable is not None:
                    task_create_tstamp = time.time()
                    for result in run_remotely.map(
                        list(stage.mappable),
                        kwargs=dict(func=stage.function, config=pipeline.config),
                    ):
                        handle_callbacks(callbacks, name, result, task_create_tstamp)
                else:
                    raise NotImplementedError()


def handle_callbacks(callbacks, array_name, result, task_create_tstamp):
    if callbacks is not None:
        task_result_tstamp = time.time()
        (
            res,
            function_start_tstamp,
            function_end_tstamp,
            peak_memory_start,
            peak_memory_end,
        ) = result
        task_stats = dict(
            task_create_tstamp=task_create_tstamp,
            function_start_tstamp=function_start_tstamp,
            function_end_tstamp=function_end_tstamp,
            task_result_tstamp=task_result_tstamp,
            peak_memory_start=peak_memory_start,
            peak_memory_end=peak_memory_end,
        )
        event = TaskEndEvent(array_name=array_name, **task_stats)
        [callback.on_task_end(event) for callback in callbacks]


class ModalDagExecutor(DagExecutor):
    """An execution engine that uses Modal."""

    def execute_dag(self, dag, callbacks=None, array_names=None, **kwargs):
        execute_dag(dag, callbacks=callbacks, array_names=array_names)



from cubed.runtime.backup import should_launch_backup


def test_should_launch_backup():
    start_times = {f"task-{task}": 0 for task in range(10)}
    end_times = {}

    # Don't launch a backup if nothing has completed yet
    assert not should_launch_backup("task-5", 7, start_times, end_times)

    end_times = {f"task-{task}": 4 for task in range(5)}

    # Don't launch a backup if not sufficiently slow (7s is not 3 times slower than 4s)
    assert not should_launch_backup("task-5", 7, start_times, end_times)

    # Don't launch a backup even if sufficiently slow (13s is 3 times slower than 4s) if not enough tasks
    assert not should_launch_backup("task-5", 13, start_times, end_times, min_tasks=20)

    # Launch a backup if sufficiently slow (13s is 3 times slower than 4s)
    assert should_launch_backup("task-5", 13, start_times, end_times)



from math import ceil, prod

from rechunker.api import _setup_rechunk

from cubed.runtime.pipeline import spec_to_pipeline


def rechunk(source, target_chunks, max_mem, target_store, temp_store=None):
    """Rechunk a Zarr array to have target_chunks.

    Parameters
    ----------
    source : Zarr array
    target_chunks : tuple
        The desired chunks of the array after rechunking.
    max_mem : int
        Maximum memory allowed for a single task of this operation, measured in bytes
    target_store : str
        Path to output Zarr store.
    temp_store : str, optional
        Path to temporary store for intermediate data.

    Returns
    -------
    pipeline:  Pipeline to run the operation
    target:  Array for the Zarr array output
    required_mem: minimum memory required per-task, in bytes
    """

    # rechunker doesn't take account of uncompressed and compressed copies of the
    # input and output array chunk/selection, so adjust appropriately
    adjusted_max_mem = max_mem / 4

    copy_specs, intermediate, target = _setup_rechunk(
        source=source,
        target_chunks=target_chunks,
        max_mem=adjusted_max_mem,
        target_store=target_store,
        temp_store=temp_store,
    )

    # source is a Zarr array, so only a single copy spec
    if len(copy_specs) != 1:  # pragma: no cover
        raise ValueError(f"Source must be a Zarr array, but was {source}")
    copy_spec = copy_specs[0]

    num_tasks = total_chunks(copy_spec.write.array.shape, copy_spec.write.chunks)
    if intermediate is not None:
        num_tasks += total_chunks(copy_spec.read.array.shape, copy_spec.read.chunks)

    return spec_to_pipeline(copy_spec, target, max_mem, num_tasks)


def total_chunks(shape, chunks):
    # cf rechunker's chunk_keys
    return prod(ceil(s / c) for s, c in zip(shape, chunks))



import numpy as np
import pytest
from numpy.testing import assert_array_equal

import cubed
import cubed.array_api as xp
from cubed.tests.utils import TaskCounter


@pytest.fixture()
def spec(tmp_path):
    return cubed.Spec(tmp_path, max_mem=100000)


def test_fusion(spec):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.negative(a)
    c = xp.astype(b, np.float32)
    d = xp.negative(c)

    assert d.plan.num_tasks(optimize_graph=False) == 12
    assert d.plan.num_tasks(optimize_graph=True) == 4

    task_counter = TaskCounter()
    result = d.compute(callbacks=[task_counter])
    assert task_counter.value == 4

    assert_array_equal(
        result,
        np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]).astype(np.float32),
    )


def test_fusion_transpose(spec):
    a = xp.asarray([[1, 2, 3], [4, 5, 6], [7, 8, 9]], chunks=(2, 2), spec=spec)
    b = xp.negative(a)
    c = xp.astype(b, np.float32)
    d = c.T

    assert d.plan.num_tasks(optimize_graph=False) == 12
    assert d.plan.num_tasks(optimize_graph=True) == 4

    task_counter = TaskCounter()
    result = d.compute(callbacks=[task_counter])
    assert task_counter.value == 4

    assert_array_equal(
        result,
        np.array([[-1, -4, -7], [-2, -5, -8], [-3, -6, -9]]).astype(np.float32),
    )


def test_no_fusion(spec):
    # b can't be fused with c because d also depends on b
    a = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    b = xp.positive(a)
    c = xp.positive(b)
    d = xp.equal(b, c)

    assert d.plan.num_tasks(optimize_graph=False) == 3
    assert d.plan.num_tasks(optimize_graph=True) == 3

    task_counter = TaskCounter()
    result = d.compute(callbacks=[task_counter])
    assert task_counter.value == 3

    assert_array_equal(result, np.ones((2, 2)))


def test_no_fusion_multiple_edges(spec):
    a = xp.ones((2, 2), chunks=(2, 2), spec=spec)
    b = xp.positive(a)
    c = xp.asarray(b)
    # b and c are the same array, so d has a single dependency
    # with multiple edges
    # this should not be fused under the current logic
    d = xp.equal(b, c)

    assert d.plan.num_tasks(optimize_graph=False) == 2
    assert d.plan.num_tasks(optimize_graph=True) == 2

    task_counter = TaskCounter()
    result = d.compute(callbacks=[task_counter])
    assert task_counter.value == 2

    assert_array_equal(result, np.full((2, 2), True))



import inspect
import tempfile
import uuid
from datetime import datetime

import fsspec
import networkx as nx
import zarr
from rechunker.types import PipelineExecutor

from cubed.primitive.blockwise import can_fuse_pipelines, fuse
from cubed.runtime.pipeline import already_computed
from cubed.utils import chunk_memory, extract_stack_summaries, join_path, memory_repr

# A unique ID with sensible ordering, used for making directory names
CONTEXT_ID = f"context-{datetime.now().strftime('%Y%m%dT%H%M%S')}-{uuid.uuid4()}"


class Plan:
    """Deferred computation plan for a graph of arrays.

    A thin wrapper around a NetworkX `MultiDiGraph`. Nodes are Zarr paths, and may
    have a `pipeline` attribute holding the pipeline to execute to generate
    the output at that path. If there is no `pipeline` attribute no computation
    is needed since the Zarr file already exists. Directed edges point towards
    the dependent Zarr path.

    Multiple edges are possible since a node (Zarr file) may be computed by
    a function with repeated inputs. For example, consider `equals` where the
    two arguments are the same array. We need to keep track of these cases, so
    we use a NetworkX `MultiDiGraph` rather than just as `DiGraph`.
    """

    def __init__(self, dag):
        self.dag = dag

    # args from pipeline onwards are omitted for creation functions when no computation is needed
    @classmethod
    def _new(
        cls,
        name,
        op_name,
        target,
        pipeline=None,
        required_mem=None,
        num_tasks=None,
        *source_arrays,
    ):
        # create an empty DAG or combine from sources
        if len(source_arrays) == 0:
            dag = nx.MultiDiGraph()
        else:
            dag = arrays_to_dag(*source_arrays)

        # add new node and edges
        frame = inspect.currentframe().f_back  # go back one in the stack
        stack_summaries = extract_stack_summaries(frame, limit=10)

        if pipeline is None:
            dag.add_node(
                name,
                name=name,
                op_name=op_name,
                target=target,
                stack_summaries=stack_summaries,
            )
        else:
            dag.add_node(
                name,
                name=name,
                op_name=op_name,
                target=target,
                stack_summaries=stack_summaries,
                pipeline=pipeline,
                required_mem=required_mem,
                num_tasks=num_tasks,
            )
        for x in source_arrays:
            if hasattr(x, "name"):
                dag.add_edge(x.name, name)

        return Plan(dag)

    @classmethod
    def arrays_to_plan(cls, *arrays):
        return Plan(arrays_to_dag(*arrays))

    def optimize(self):
        # note there is no need to prune the dag, since the way it is built
        # ensures that only the transitive dependencies of the target arrays are included

        # fuse map blocks
        dag = self.dag.copy()
        nodes = {n: d for (n, d) in dag.nodes(data=True)}

        def can_fuse(n):
            # node must have a single predecessor
            #   - not multiple edges pointing to a single predecessor
            # node must be the single successor to the predecessor
            # and both must have pipelines that can be fused
            if dag.in_degree(n) != 1:
                return False
            pre = next(dag.predecessors(n))
            if dag.out_degree(pre) != 1:
                return False
            if "pipeline" not in nodes[pre] or "pipeline" not in nodes[n]:
                return False
            return can_fuse_pipelines(nodes[pre]["pipeline"], nodes[n]["pipeline"])

        for n in list(dag.nodes()):
            if can_fuse(n):
                pre = next(dag.predecessors(n))
                pipeline = fuse(nodes[pre]["pipeline"], nodes[n]["pipeline"])
                nodes[n]["pipeline"] = pipeline
                assert nodes[n]["target"] == pipeline.target_array

                for p in dag.predecessors(pre):
                    dag.add_edge(p, n)
                dag.remove_node(pre)

        return Plan(dag)

    def execute(
        self,
        executor=None,
        callbacks=None,
        optimize_graph=True,
        array_names=None,
        **kwargs,
    ):
        dag = self.optimize().dag if optimize_graph else self.dag.copy()

        if isinstance(executor, PipelineExecutor):
            while len(dag) > 0:
                # Find nodes (and their pipelines) that have no dependencies
                no_dep_nodes = [x for x in dag.nodes() if dag.in_degree(x) == 0]
                pipelines = [
                    p
                    for (n, p) in nx.get_node_attributes(dag, "pipeline").items()
                    if n in no_dep_nodes
                ]

                # and execute them in parallel
                if len(pipelines) > 0:
                    plan = executor.pipelines_to_plan(pipelines)
                    executor.execute_plan(plan, **kwargs)

                # Remove them from the DAG, and repeat
                dag.remove_nodes_from(no_dep_nodes)

        else:
            if callbacks is not None:
                [callback.on_compute_start(dag) for callback in callbacks]
            executor.execute_dag(
                dag, callbacks=callbacks, array_names=array_names, **kwargs
            )
            if callbacks is not None:
                [callback.on_compute_end(dag) for callback in callbacks]

    def num_tasks(self, optimize_graph=True):
        """Return the number of tasks needed to execute this plan."""
        dag = self.optimize().dag if optimize_graph else self.dag.copy()
        tasks = 0
        for _, node in visit_nodes(dag):
            pipeline = node["pipeline"]
            for stage in pipeline.stages:
                if stage.mappable is not None:
                    tasks += len(list(stage.mappable))
                else:
                    tasks += 1
        return tasks

    def max_required_mem(self, optimize_graph=True):
        """Return the maximum required memory (for a task) to execute this plan."""
        dag = self.optimize().dag if optimize_graph else self.dag.copy()
        required_mem_values = [
            node["pipeline"].required_mem for _, node in visit_nodes(dag)
        ]
        return max(required_mem_values) if len(required_mem_values) > 0 else 0

    def visualize(
        self, filename="cubed", format=None, rankdir="TB", optimize_graph=True
    ):
        dag = self.optimize().dag if optimize_graph else self.dag.copy()
        dag.graph["graph"] = {
            "rankdir": rankdir,
            "label": (
                f"num tasks: {self.num_tasks(optimize_graph=optimize_graph)}\n"
                f"max required memory: {memory_repr(self.max_required_mem(optimize_graph=optimize_graph))}"
            ),
            "labelloc": "bottom",
            "labeljust": "left",
            "fontsize": "10",
        }
        dag.graph["node"] = {"fontname": "helvetica", "shape": "box", "fontsize": "10"}
        for (n, d) in dag.nodes(data=True):
            if d["op_name"] == "blockwise":
                d["style"] = "filled"
                d["fillcolor"] = "#dcbeff"
                op_name_summary = "(bw)"
            elif d["op_name"] == "rechunk":
                d["style"] = "filled"
                d["fillcolor"] = "#aaffc3"
                op_name_summary = "(rc)"
            else:  # creation function
                op_name_summary = ""
            target = d["target"]
            chunkmem = memory_repr(chunk_memory(target.dtype, target.chunks))
            tooltip = (
                f"shape: {target.shape}\n"
                f"chunks: {target.chunks}\n"
                f"dtype: {target.dtype}\n"
                f"chunk memory: {chunkmem}\n"
            )
            if "pipeline" in d:
                pipeline = d["pipeline"]
                tooltip += f"\ntask memory: {memory_repr(pipeline.required_mem)}"
                tooltip += f"\ntasks: {pipeline.num_tasks}"
            if "stack_summaries" in d:
                # add call stack information
                stack_summaries = d["stack_summaries"]

                first_cubed_i = min(
                    i for i, s in enumerate(stack_summaries) if s.is_cubed()
                )
                first_cubed_summary = stack_summaries[first_cubed_i]
                caller_summary = stack_summaries[first_cubed_i - 1]

                d["label"] = f"{n}\n{first_cubed_summary.name} {op_name_summary}"

                calls = " -> ".join(
                    [s.name for s in stack_summaries if not s.is_on_python_lib_path()]
                )

                line = f"{caller_summary.lineno} in {caller_summary.name}"

                tooltip += f"\ncalls: {calls}"
                tooltip += f"\nline: {line}"
                del d["stack_summaries"]

            d["tooltip"] = tooltip

            # remove pipeline attribute since it is a long string that causes graphviz to fail
            if "pipeline" in d:
                del d["pipeline"]
            if "target" in d:
                del d["target"]
            if "name" in d:  # pydot already has name
                del d["name"]
        gv = nx.drawing.nx_pydot.to_pydot(dag)
        if format is None:
            format = "svg"
        full_filename = f"{filename}.{format}"
        gv.write(full_filename, format=format)

        try:  # pragma: no cover
            import IPython.display as display

            if format == "svg":
                return display.SVG(filename=full_filename)
        except ImportError:
            # Can't return a display object if no IPython.
            pass
        return None


def arrays_to_dag(*arrays):
    from .array import check_array_specs

    check_array_specs(arrays)
    dags = [x.plan.dag for x in arrays if hasattr(x, "plan")]
    return nx.compose_all(dags)


def arrays_to_plan(*arrays):
    plans = [x.plan for x in arrays if hasattr(x, "plan")]
    if len(plans) == 0:
        raise ValueError(f"No plans found for arrays: {arrays}")
    return plans[0].arrays_to_plan(*arrays)


def new_temp_path(name, suffix, spec=None):
    work_dir = spec.work_dir if spec is not None else None
    if work_dir is None:
        work_dir = tempfile.gettempdir()
    context_dir = join_path(work_dir, CONTEXT_ID)
    return join_path(context_dir, f"{name}{suffix}")


def new_temp_store(name, spec=None):
    zarr_path = new_temp_path(name, ".zarr", spec)
    return fsspec.get_mapper(zarr_path)


def new_temp_zarr(shape, dtype, chunksize, name=None, spec=None):
    # open a new temporary zarr array for writing
    store = new_temp_store(name=name, spec=spec)
    target = zarr.open(store, mode="w-", shape=shape, dtype=dtype, chunks=chunksize)
    return target


def visit_nodes(dag):
    """Return a generator that visits the nodes in the DAG in topological order."""
    nodes = {n: d for (n, d) in dag.nodes(data=True)}
    for name in list(nx.topological_sort(dag)):
        if already_computed(nodes[name]):
            continue
        yield name, nodes[name]


def visit_node_generations(dag):
    """Return a generator that visits the nodes in the DAG in groups of topological generations."""
    nodes = {n: d for (n, d) in dag.nodes(data=True)}
    for names in nx.topological_generations(dag):
        gen = [
            (name, nodes[name]) for name in names if not already_computed(nodes[name])
        ]
        if len(gen) > 0:
            yield gen



from typing import Iterable

import numpy as np
import zarr
from rechunker.executors.python import PythonPipelineExecutor

from cubed.core.array import Callback
from cubed.runtime.executors.python import PythonDagExecutor
from cubed.runtime.executors.python_async import AsyncPythonDagExecutor

LITHOPS_LOCAL_CONFIG = {"lithops": {"backend": "localhost", "storage": "localhost"}}

ALL_EXECUTORS = [
    PythonPipelineExecutor(),
    PythonDagExecutor(),
    AsyncPythonDagExecutor(),
]

# don't run all tests on every executor as it's too slow, so just have a subset
MAIN_EXECUTORS = [PythonPipelineExecutor(), PythonDagExecutor()]

try:
    from cubed.runtime.executors.beam import BeamDagExecutor, BeamPipelineExecutor

    ALL_EXECUTORS.append(BeamDagExecutor())
    ALL_EXECUTORS.append(BeamPipelineExecutor())

    MAIN_EXECUTORS.append(BeamDagExecutor())
except ImportError:
    pass

try:
    from cubed.runtime.executors.lithops import (
        LithopsDagExecutor,
        LithopsPipelineExecutor,
    )

    ALL_EXECUTORS.append(LithopsDagExecutor(config=LITHOPS_LOCAL_CONFIG))
    ALL_EXECUTORS.append(LithopsPipelineExecutor(config=LITHOPS_LOCAL_CONFIG))

    MAIN_EXECUTORS.append(LithopsDagExecutor(config=LITHOPS_LOCAL_CONFIG))
except ImportError:
    pass

MODAL_EXECUTORS = []

try:
    from cubed.runtime.executors.modal import ModalDagExecutor
    from cubed.runtime.executors.modal_async import AsyncModalDagExecutor

    MODAL_EXECUTORS.append(AsyncModalDagExecutor())
    MODAL_EXECUTORS.append(ModalDagExecutor())
except ImportError:
    pass


class TaskCounter(Callback):
    def on_compute_start(self, dag):
        self.value = 0

    def on_task_end(self, event):
        if event.task_create_tstamp is not None:
            assert (
                event.task_result_tstamp
                >= event.function_end_tstamp
                >= event.function_start_tstamp
                >= event.task_create_tstamp
                > 0
            )
        self.value += event.num_tasks


def create_zarr(a, /, store, *, dtype=None, chunks=None):
    # from dask.asarray
    if not isinstance(getattr(a, "shape", None), Iterable):
        # ensure blocks are arrays
        a = np.asarray(a, dtype=dtype)
    if dtype is None:
        dtype = a.dtype

    # write to zarr
    za = zarr.open(store, mode="w", shape=a.shape, dtype=dtype, chunks=chunks)
    za[:] = a
    return za


def execute_pipeline(pipeline, executor):
    """Executes a pipeline"""
    plan = executor.pipelines_to_plan([pipeline])
    executor.execute_plan(plan)



import numpy as np

from cubed.array_api.creation_functions import asarray
from cubed.core import reduction


def all(x, /, *, axis=None, keepdims=False):
    if x.size == 0:
        return asarray(True, dtype=x.dtype)
    return reduction(x, np.all, axis=axis, dtype=bool, keepdims=keepdims)


def any(x, /, *, axis=None, keepdims=False):
    if x.size == 0:
        return asarray(False, dtype=x.dtype)
    return reduction(x, np.any, axis=axis, dtype=bool, keepdims=keepdims)



from typing import TYPE_CHECKING, Iterable, List

import numpy as np
import zarr
from dask.array.core import normalize_chunks
from zarr.util import normalize_shape

from cubed.core import Plan, gensym, map_blocks, new_temp_store, new_temp_zarr
from cubed.core.ops import map_direct
from cubed.utils import to_chunksize

if TYPE_CHECKING:
    from .array_object import Array


def arange(
    start, /, stop=None, step=1, *, dtype=None, device=None, chunks="auto", spec=None
) -> "Array":
    if stop is None:
        start, stop = 0, start
    num = int(max(np.ceil((stop - start) / step), 0))
    if dtype is None:
        dtype = np.arange(start, stop, step * num if num else step).dtype
    chunks = normalize_chunks(chunks, shape=(num,), dtype=dtype)
    chunksize = chunks[0][0]
    numblocks = len(chunks[0])
    # create small array of block numbers
    out = asarray(np.arange(numblocks), chunks=(1,), spec=spec)
    # then map each block to partial arange
    out = map_blocks(
        _arange,
        out,
        dtype=dtype,
        chunks=chunks,
        size=chunksize,
        start=start,
        stop=stop,
        step=step,
    )
    return out


def _arange(a, size, start, stop, step):
    i = a[0]
    blockstart = start + (i * size * step)
    blockstop = start + ((i + 1) * size * step)
    return np.arange(blockstart, min(blockstop, stop), step)


def asarray(
    obj, /, *, dtype=None, device=None, copy=None, chunks="auto", spec=None
) -> "Array":
    a = obj
    from cubed.array_api.array_object import Array

    # from dask.asarray
    if isinstance(a, Array):
        return a
    elif type(a).__module__.split(".")[0] == "xarray" and hasattr(
        a, "data"
    ):  # pragma: no cover
        return asarray(a.data)
    elif not isinstance(getattr(a, "shape", None), Iterable):
        # ensure blocks are arrays
        a = np.asarray(a, dtype=dtype)
    if dtype is None:
        dtype = a.dtype

    # write to zarr
    chunksize = to_chunksize(normalize_chunks(chunks, shape=a.shape, dtype=dtype))
    name = gensym()
    target = new_temp_zarr(a.shape, dtype, chunksize, name=name, spec=spec)
    if a.size > 0:
        target[...] = a

    plan = Plan._new(name, "asarray", target)
    return Array(name, target, spec, plan)


def empty(shape, *, dtype=None, device=None, chunks="auto", spec=None) -> "Array":
    if dtype is None:
        dtype = np.float64
    return full(shape, None, dtype=dtype, device=device, chunks=chunks, spec=spec)


def empty_like(x, /, *, dtype=None, device=None, chunks=None, spec=None) -> "Array":
    return empty(**_like_args(x, dtype, device, chunks, spec))


def eye(
    n_rows, n_cols=None, /, *, k=0, dtype=None, device=None, chunks="auto", spec=None
) -> "Array":
    if n_cols is None:
        n_cols = n_rows
    if dtype is None:
        dtype = np.float64

    shape = (n_rows, n_cols)
    chunks = normalize_chunks(chunks, shape=shape, dtype=dtype)
    chunksize = to_chunksize(chunks)[0]

    return map_direct(
        _eye,
        shape=shape,
        dtype=dtype,
        chunks=chunks,
        extra_required_mem=0,
        spec=spec,
        k=k,
        chunksize=chunksize,
    )


def _eye(x, *arrays, k=None, chunksize=None, block_id=None):
    i, j = block_id
    bk = (j - i) * chunksize
    if bk - chunksize <= k <= bk + chunksize:
        return np.eye(x.shape[0], x.shape[1], k=k - bk, dtype=x.dtype)
    else:
        return np.zeros_like(x)


def full(
    shape, fill_value, *, dtype=None, device=None, chunks="auto", spec=None
) -> "Array":
    # write to zarr
    # note that write_empty_chunks=False means no chunks are written to disk, so it is very efficient to create large arrays
    shape = normalize_shape(shape)
    if dtype is None:
        if isinstance(fill_value, int):
            dtype = np.int64
        elif isinstance(fill_value, float):
            dtype = np.float64
        elif isinstance(fill_value, bool):
            dtype = np.bool_
        else:
            raise TypeError("Invalid input to full")
    chunksize = to_chunksize(normalize_chunks(chunks, shape=shape, dtype=dtype))
    name = gensym()
    store = new_temp_store(name=name, spec=spec)
    target = zarr.full(
        shape,
        fill_value,
        store=store,
        dtype=dtype,
        chunks=chunksize,
        write_empty_chunks=False,
    )

    from .array_object import Array

    plan = Plan._new(name, "full", target)
    return Array(name, target, spec, plan)


def full_like(
    x, /, fill_value, *, dtype=None, device=None, chunks=None, spec=None
) -> "Array":
    return full(fill_value=fill_value, **_like_args(x, dtype, device, chunks, spec))


def linspace(
    start,
    stop,
    /,
    num,
    *,
    dtype=None,
    device=None,
    endpoint=True,
    chunks="auto",
    spec=None,
) -> "Array":
    range_ = stop - start
    div = (num - 1) if endpoint else num
    if div == 0:
        div = 1
    step = float(range_) / div
    shape = (num,)
    if dtype is None:
        dtype = np.float64
    chunks = normalize_chunks(chunks, shape=shape, dtype=dtype)
    chunksize = chunks[0][0]

    if num == 0:
        return asarray(0.0, dtype=dtype, spec=spec)

    return map_direct(
        _linspace,
        shape=shape,
        dtype=dtype,
        chunks=chunks,
        extra_required_mem=0,
        spec=spec,
        size=chunksize,
        start=start,
        step=step,
        endpoint=endpoint,
        linspace_dtype=dtype,
    )


def _linspace(x, *arrays, size, start, step, endpoint, linspace_dtype, block_id=None):
    bs = x.shape[0]
    i = block_id[0]
    adjusted_bs = bs - 1 if endpoint else bs
    blockstart = start + (i * size * step)
    blockstop = blockstart + (adjusted_bs * step)
    return np.linspace(
        blockstart, blockstop, bs, endpoint=endpoint, dtype=linspace_dtype
    )


def meshgrid(*arrays, indexing="xy") -> List["Array"]:
    if len({a.dtype for a in arrays}) > 1:
        raise ValueError("meshgrid inputs must all have the same dtype")

    from cubed.array_api.manipulation_functions import broadcast_arrays

    # based on dask
    if indexing not in ("ij", "xy"):
        raise ValueError("`indexing` must be `'ij'` or `'xy'`")

    arrs = list(arrays)
    if indexing == "xy" and len(arrs) > 1:
        arrs[0], arrs[1] = arrs[1], arrs[0]

    grid = []
    for i in range(len(arrs)):
        s = len(arrs) * [None]
        s[i] = slice(None)  # type: ignore[call-overload]

        r = arrs[i][tuple(s)]

        grid.append(r)

    grid = list(broadcast_arrays(*grid))

    if indexing == "xy" and len(arrs) > 1:
        grid[0], grid[1] = grid[1], grid[0]

    return grid


def ones(shape, *, dtype=None, device=None, chunks="auto", spec=None) -> "Array":
    if dtype is None:
        dtype = np.float64
    return full(shape, 1, dtype=dtype, device=device, chunks=chunks, spec=spec)


def ones_like(x, /, *, dtype=None, device=None, chunks=None, spec=None) -> "Array":
    return ones(**_like_args(x, dtype, device, chunks, spec))


def tril(x, /, *, k=0) -> "Array":
    from cubed.array_api.searching_functions import where

    if x.ndim < 2:
        raise ValueError("x must be at least 2-dimensional for tril")

    mask = _tri_mask(x.shape[-2], x.shape[-1], k, x.chunks[-2:], x.spec)
    return where(mask, x, zeros_like(x))


def triu(x, /, *, k=0) -> "Array":
    from cubed.array_api.searching_functions import where

    if x.ndim < 2:
        raise ValueError("x must be at least 2-dimensional for triu")

    mask = _tri_mask(x.shape[-2], x.shape[-1], k - 1, x.chunks[-2:], x.spec)
    return where(mask, zeros_like(x), x)


def _tri_mask(N, M, k, chunks, spec):
    from cubed.array_api.elementwise_functions import greater_equal
    from cubed.array_api.manipulation_functions import expand_dims

    # based on dask
    chunks = normalize_chunks(chunks, shape=(N, M))

    # TODO: use min_int for arange dtype
    m = greater_equal(
        expand_dims(arange(N, chunks=chunks[0][0], spec=spec), axis=1),
        arange(-k, M - k, chunks=chunks[1][0], spec=spec),
    )

    return m


def zeros(shape, *, dtype=None, device=None, chunks="auto", spec=None) -> "Array":
    if dtype is None:
        dtype = np.float64
    return full(shape, 0, dtype=dtype, device=device, chunks=chunks, spec=spec)


def zeros_like(x, /, *, dtype=None, device=None, chunks=None, spec=None) -> "Array":
    return zeros(**_like_args(x, dtype, device, chunks, spec))


def _like_args(x, dtype=None, device=None, chunks=None, spec=None):
    if dtype is None:
        dtype = x.dtype
    if chunks is None:
        chunks = x.chunks
    if spec is None:
        spec = x.spec
    return dict(shape=x.shape, dtype=dtype, device=device, chunks=chunks, spec=spec)



import asyncio
import copy
import os
import time
from asyncio.exceptions import TimeoutError

import modal
import modal.aio
from modal.exception import ConnectionError
from tenacity import retry, retry_if_exception_type, stop_after_attempt

from cubed.core.array import TaskEndEvent
from cubed.core.plan import visit_nodes
from cubed.runtime.backup import should_launch_backup
from cubed.runtime.types import DagExecutor
from cubed.utils import peak_memory

async_stub = modal.aio.AioStub()

requirements_file = os.getenv("CUBED_MODAL_REQUIREMENTS_FILE")

if requirements_file:
    image = modal.DebianSlim().pip_install_from_requirements(requirements_file)
else:
    image = modal.DebianSlim().pip_install(
        [
            "dask[array]",
            "fsspec",
            "networkx",
            "pytest-mock",  # TODO: only needed for tests
            "rechunker",
            "s3fs",
            "tenacity",
            "zarr",
        ]
    )


@async_stub.generator(
    image=image, secret=modal.ref("my-aws-secret"), memory=2000, retries=2
)
async def async_run_remotely(input, func=None, config=None):
    print(f"running remotely on {input}")
    peak_memory_start = peak_memory()
    function_start_tstamp = time.time()
    result = func(input, config=config)
    function_end_tstamp = time.time()
    peak_memory_end = peak_memory()
    yield (
        result,
        function_start_tstamp,
        function_end_tstamp,
        peak_memory_start,
        peak_memory_end,
    )


# We need map_unordered for the use_backups implementation
async def map_unordered(
    app_function, input, max_failures=3, use_backups=False, **kwargs
):
    """
    Apply a function to items of an input list, yielding results as they are completed
    (which may be different to the input order).

    :param app_function: The Modal function to map over the data.
    :param input: An iterable of input data.
    :param max_failures: The number of task failures to allow before raising an exception.
    :param use_backups: Whether to launch backup tasks to mitigate against slow-running tasks.
    :param kwargs: Keyword arguments to pass to the function.

    :return: Function values (and optionally stats) as they are completed, not necessarily in the input order.
    """
    if not use_backups:
        async for result in app_function.map(input, kwargs=kwargs):
            yield result
        return

    tasks = {asyncio.ensure_future(app_function(i, **kwargs)): i for i in input}
    pending = set(tasks.keys())
    t = time.monotonic()
    start_times = {f: t for f in pending}
    end_times = {}
    backups = {}

    while pending:
        finished, pending = await asyncio.wait(
            pending, return_when=asyncio.FIRST_COMPLETED, timeout=2
        )
        # print("finished", finished)
        # print("pending", pending)

        for task in finished:
            if task.exception():
                raise task.exception()
            end_times[task] = time.monotonic()
            yield task.result()

            # remove any backup task
            if use_backups:
                backup = backups.get(task, None)
                if backup:
                    pending.remove(backup)
                    del backups[task]
                    del backups[backup]

        if use_backups:
            now = time.monotonic()
            for task in copy.copy(pending):
                if task not in backups and should_launch_backup(
                    task, now, start_times, end_times
                ):
                    # launch backup task
                    i = tasks[task]
                    new_task = asyncio.ensure_future(app_function(i, **kwargs))
                    tasks[new_task] = i
                    start_times[new_task] = time.monotonic()
                    pending.add(new_task)
                    backups[task] = new_task
                    backups[new_task] = task


# This just retries the initial connection attempt, not the function calls
@retry(
    retry=retry_if_exception_type((TimeoutError, ConnectionError)),
    stop=stop_after_attempt(3),
)
async def async_execute_dag(dag, callbacks=None, array_names=None, **kwargs):
    async with async_stub.run():
        for name, node in visit_nodes(dag):
            pipeline = node["pipeline"]

            for stage in pipeline.stages:
                if stage.mappable is not None:
                    task_create_tstamp = time.time()
                    async for result in map_unordered(
                        async_run_remotely,
                        list(stage.mappable),
                        func=stage.function,
                        config=pipeline.config,
                        **kwargs,
                    ):
                        handle_callbacks(callbacks, name, result, task_create_tstamp)
                else:
                    raise NotImplementedError()


def handle_callbacks(callbacks, array_name, result, task_create_tstamp):
    if callbacks is not None:
        task_result_tstamp = time.time()
        (
            res,
            function_start_tstamp,
            function_end_tstamp,
            peak_memory_start,
            peak_memory_end,
        ) = result
        task_stats = dict(
            task_create_tstamp=task_create_tstamp,
            function_start_tstamp=function_start_tstamp,
            function_end_tstamp=function_end_tstamp,
            task_result_tstamp=task_result_tstamp,
            peak_memory_start=peak_memory_start,
            peak_memory_end=peak_memory_end,
        )
        event = TaskEndEvent(array_name=array_name, **task_stats)
        [callback.on_task_end(event) for callback in callbacks]


class AsyncModalDagExecutor(DagExecutor):
    """An execution engine that uses Modal's async API."""

    def execute_dag(self, dag, callbacks=None, array_names=None, **kwargs):
        asyncio.run(
            async_execute_dag(
                dag, callbacks=callbacks, array_names=array_names, **kwargs
            )
        )



# Suppress numpy.array_api experimental warning
import sys
import warnings

if not sys.warnoptions:
    warnings.filterwarnings("ignore", category=UserWarning)

from importlib.metadata import version as _version

try:
    __version__ = _version("cubed")
except Exception:  # pragma: no cover
    __version__ = "unknown"

from .array_api import Array
from .core.array import (
    Callback,
    Spec,
    TaskEndEvent,
    compute,
    measure_reserved_memory,
    visualize,
)
from .core.ops import from_array, from_zarr, map_blocks, store, to_zarr

__all__ = [
    "__version__",
    "Callback",
    "Array",
    "Spec",
    "TaskEndEvent",
    "compute",
    "from_array",
    "from_zarr",
    "map_blocks",
    "measure_reserved_memory",
    "store",
    "to_zarr",
    "visualize",
]



import time
from dataclasses import asdict
from pathlib import Path

import pandas as pd

from cubed.core.array import Callback
from cubed.core.plan import visit_nodes


class HistoryCallback(Callback):
    def on_compute_start(self, dag):
        plan = []
        for name, node in visit_nodes(dag):
            pipeline = node["pipeline"]
            plan.append(
                dict(
                    array_name=name,
                    op_name=node["op_name"],
                    required_mem=pipeline.required_mem,
                    num_tasks=pipeline.num_tasks,
                )
            )

        self.plan = plan
        self.stats = []

    def on_task_end(self, event):
        self.stats.append(asdict(event))

    def on_compute_end(self, dag):
        plan_df = pd.DataFrame(self.plan)
        stats_df = pd.DataFrame(self.stats)
        Path("history").mkdir(exist_ok=True)
        id = int(time.time())
        self.plan_df_path = Path(f"history/plan-{id}.csv")
        self.stats_df_path = Path(f"history/stats-{id}.csv")
        plan_df.to_csv(self.plan_df_path, index=False)
        stats_df.to_csv(self.stats_df_path, index=False)



import random as pyrandom

import numpy as np
import numpy.array_api as nxp
from dask.array.core import normalize_chunks
from numpy.random import Generator, Philox
from zarr.util import normalize_shape

from cubed.core.ops import map_direct


def random(size, *, chunks=None, spec=None):
    """Return random floats in the half-open interval [0.0, 1.0)."""
    shape = normalize_shape(size)
    dtype = nxp.float64
    chunks = normalize_chunks(chunks, shape=shape, dtype=dtype)
    numblocks = tuple(map(len, chunks))
    root_seed = pyrandom.getrandbits(128)

    # no extra memory required since input is an empty array whose
    # memory is never allocated, see https://pythonspeed.com/articles/measuring-memory-python/#phantom-memory
    extra_required_mem = 0

    return map_direct(
        _random,
        shape=shape,
        dtype=dtype,
        chunks=chunks,
        extra_required_mem=extra_required_mem,
        spec=spec,
        numblocks=numblocks,
        root_seed=root_seed,
    )


def _random(x, *arrays, numblocks=None, root_seed=None, block_id=None):
    stream_id = block_id_to_offset(block_id, numblocks)
    rg = Generator(Philox(key=root_seed + stream_id))
    return rg.random(x.shape)


def block_id_to_offset(block_id, numblocks):
    arr = np.empty(numblocks, dtype=np.int8)
    return sum((np.array(block_id) * arr.strides).tolist())



import inspect

import numpy as np
import pytest

from cubed.utils import (
    chunk_memory,
    extract_stack_summaries,
    join_path,
    memory_repr,
    peak_memory,
    to_chunksize,
)


def test_chunk_memory():
    assert chunk_memory(int, (3,)) == 24
    assert chunk_memory(np.int64, (3,)) == 24
    assert chunk_memory(np.int32, (3,)) == 12
    assert chunk_memory(np.int32, (3, 5)) == 60
    assert chunk_memory(np.int32, (0,)) == 0


def test_to_chunksize():
    assert to_chunksize(((3, 3, 3, 1),)) == (3,)
    with pytest.raises(ValueError):
        to_chunksize(((3, 2, 3, 3, 1),))


def test_join_path():
    assert join_path("http://host/path", "subpath") == "http://host/path/subpath"
    assert join_path("http://host/path/", "subpath") == "http://host/path/subpath"
    assert (
        join_path("http://host/path?a=b", "subpath") == "http://host/path/subpath?a=b"
    )
    assert (
        join_path("http://host/path/?a=b", "subpath") == "http://host/path/subpath?a=b"
    )
    assert join_path("http://host/path#a", "subpath") == "http://host/path/subpath#a"
    assert join_path("s3://host/path", "subpath") == "s3://host/path/subpath"
    assert join_path("relative_path/path", "subpath") == "relative_path/path/subpath"
    assert join_path("/absolute_path/path", "subpath") == "/absolute_path/path/subpath"
    assert (
        join_path("http://host/a%20path", "subpath") == "http://host/a%20path/subpath"
    )
    assert join_path("http://host/a path", "subpath") == "http://host/a%20path/subpath"


def test_memory_repr():
    assert memory_repr(0) == "0 bytes"
    assert memory_repr(1) == "1 bytes"
    assert memory_repr(999) == "999 bytes"
    assert memory_repr(1_000) == "1.0 KB"
    assert memory_repr(9_999) == "10.0 KB"
    assert memory_repr(1_000_000) == "1.0 MB"
    assert memory_repr(1_000_000_000_000_000) == "1.0 PB"


def test_peak_memory():
    assert peak_memory() > 0


def test_extract_stack_summaries():
    frame = inspect.currentframe()
    stack_summaries = extract_stack_summaries(frame)
    assert stack_summaries[-1].name == "test_extract_stack_summaries"
    assert stack_summaries[-1].module == "cubed.tests.test_utils"
    assert not stack_summaries[-1].is_cubed()



# flake8: noqa
from .array import (
    Callback,
    CoreArray,
    Spec,
    TaskEndEvent,
    compute,
    gensym,
    measure_reserved_memory,
    visualize,
)
from .ops import (
    blockwise,
    elemwise,
    from_array,
    from_zarr,
    map_blocks,
    rechunk,
    reduction,
    squeeze,
    store,
    to_zarr,
    unify_chunks,
)
from .plan import Plan, new_temp_store, new_temp_zarr



import random

import pytest
from numpy.testing import assert_array_equal

import cubed
import cubed.array_api as xp
import cubed.random
from cubed.tests.utils import MAIN_EXECUTORS


@pytest.fixture()
def spec(tmp_path):
    return cubed.Spec(tmp_path, max_mem=100000)


@pytest.fixture(scope="module", params=MAIN_EXECUTORS)
def executor(request):
    return request.param


def test_random(spec, executor):
    a = cubed.random.random((10, 10), chunks=(4, 5), spec=spec)

    assert a.shape == (10, 10)
    assert a.chunks == ((4, 4, 2), (5, 5))

    x = set(a.compute(executor=executor).flat)
    assert len(x) > 90


def test_random_add(spec, executor):
    a = cubed.random.random((10, 10), chunks=(5, 5), spec=spec)
    b = cubed.random.random((10, 10), chunks=(5, 5), spec=spec)

    c = xp.add(a, b)

    x = set(c.compute(executor=executor).flat)
    assert len(x) > 90


def test_random_seed(spec, executor):
    random.seed(42)
    a = cubed.random.random((10, 10), chunks=(5, 5), spec=spec)
    a_result = a.compute(executor=executor)

    random.seed(42)
    b = cubed.random.random((10, 10), chunks=(5, 5), spec=spec)
    b_result = b.compute(executor=executor)

    assert_array_equal(a_result, b_result)



import numpy as np
import pytest
import zarr
from dask.blockwise import make_blockwise_graph
from numpy.testing import assert_array_equal
from rechunker.executors.python import PythonPipelineExecutor

from cubed.primitive.blockwise import blockwise, make_blockwise_function
from cubed.tests.utils import create_zarr, execute_pipeline


@pytest.fixture(scope="module", params=[PythonPipelineExecutor()])
def executor(request):
    return request.param


def test_blockwise(tmp_path, executor):
    source1 = create_zarr(
        [0, 1, 2], dtype=int, chunks=2, store=tmp_path / "source1.zarr"
    )
    source2 = create_zarr(
        [10, 50, 100], dtype=int, chunks=2, store=tmp_path / "source2.zarr"
    )
    max_mem = 1000
    target_store = tmp_path / "target.zarr"

    pipeline = blockwise(
        np.outer,
        "ij",
        source1,
        "i",
        source2,
        "j",
        max_mem=max_mem,
        target_store=target_store,
        shape=(3, 3),
        dtype=int,
        chunks=(2, 2),
    )

    assert pipeline.target_array.shape == (3, 3)
    assert pipeline.target_array.dtype == int
    assert pipeline.target_array.chunks == (2, 2)

    itemsize = np.dtype(int).itemsize
    assert pipeline.required_mem == (
        (itemsize * 2)  # source1 compressed chunk
        + (itemsize * 2)  # source1 uncompressed chunk
        + (itemsize * 2)  # source2 compressed chunk
        + (itemsize * 2)  # source2 uncompressed chunk
        + (itemsize * 2 * 2)  # output compressed chunk
        + (itemsize * 2 * 2)  # output uncompressed chunk
    )

    assert pipeline.num_tasks == 4

    execute_pipeline(pipeline, executor=executor)

    res = zarr.open(target_store)
    assert_array_equal(res[:], np.outer([0, 1, 2], [10, 50, 100]))


def _permute_dims(x, /, axes, max_mem, target_store):
    # From dask transpose
    if axes:
        if len(axes) != x.ndim:
            raise ValueError("axes don't match array")
    else:
        axes = tuple(range(x.ndim))[::-1]
    axes = tuple(d + x.ndim if d < 0 else d for d in axes)
    return blockwise(
        np.transpose,
        axes,
        x,
        tuple(range(x.ndim)),
        max_mem=max_mem,
        target_store=target_store,
        shape=x.shape,
        dtype=x.dtype,
        chunks=x.chunks,
        axes=axes,
    )


def test_blockwise_with_args(tmp_path, executor):
    source = create_zarr(
        [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
        dtype=int,
        chunks=(2, 2),
        store=tmp_path / "source.zarr",
    )
    max_mem = 1000
    target_store = tmp_path / "target.zarr"

    pipeline = _permute_dims(
        source, axes=(1, 0), max_mem=max_mem, target_store=target_store
    )

    assert pipeline.target_array.shape == (3, 3)
    assert pipeline.target_array.dtype == int
    assert pipeline.target_array.chunks == (2, 2)

    itemsize = np.dtype(int).itemsize
    assert pipeline.required_mem == (
        (itemsize * 2 * 2)  # source compressed chunk
        + (itemsize * 2 * 2)  # source uncompressed chunk
        + (itemsize * 2 * 2)  # output compressed chunk
        + (itemsize * 2 * 2)  # output uncompressed chunk
    )

    assert pipeline.num_tasks == 4

    execute_pipeline(pipeline, executor=executor)

    res = zarr.open(target_store)
    assert_array_equal(
        res[:], np.transpose(np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]), axes=(1, 0))
    )


def test_blockwise_max_mem_exceeded(tmp_path):
    source1 = create_zarr(
        [0, 1, 2], dtype=int, chunks=2, store=tmp_path / "source1.zarr"
    )
    source2 = create_zarr(
        [10, 50, 100], dtype=int, chunks=2, store=tmp_path / "source2.zarr"
    )
    max_mem = 100
    target_store = tmp_path / "target.zarr"

    with pytest.raises(
        ValueError, match=r"Blockwise memory \(\d+\) exceeds max_mem \(100\)"
    ):
        blockwise(
            np.outer,
            "ij",
            source1,
            "i",
            source2,
            "j",
            max_mem=max_mem,
            target_store=target_store,
            shape=(3, 3),
            dtype=int,
            chunks=(2, 2),
        )


def test_make_blockwise_function_map():
    func = lambda x: 0

    block_fn = make_blockwise_function(
        func, "z", "ij", "x", "ij", numblocks={"x": (2, 2)}
    )

    graph = make_blockwise_graph(func, "z", "ij", "x", "ij", numblocks={"x": (2, 2)})
    check_consistent_with_graph(block_fn, graph)


def test_make_blockwise_function_elemwise():
    func = lambda x: 0

    block_fn = make_blockwise_function(
        func, "z", "ij", "x", "ij", "y", "ij", numblocks={"x": (2, 2), "y": (2, 2)}
    )

    graph = make_blockwise_graph(
        func, "z", "ij", "x", "ij", "y", "ij", numblocks={"x": (2, 2), "y": (2, 2)}
    )
    check_consistent_with_graph(block_fn, graph)


def test_make_blockwise_function_flip():
    func = lambda x: 0

    block_fn = make_blockwise_function(
        func, "z", "ij", "x", "ij", "y", "ji", numblocks={"x": (2, 2), "y": (2, 2)}
    )

    graph = make_blockwise_graph(
        func, "z", "ij", "x", "ij", "y", "ji", numblocks={"x": (2, 2), "y": (2, 2)}
    )
    check_consistent_with_graph(block_fn, graph)


def test_make_blockwise_function_contract():
    func = lambda x: 0

    block_fn = make_blockwise_function(
        func, "z", "ik", "x", "ij", "y", "jk", numblocks={"x": (2, 2), "y": (2, 2)}
    )

    graph = make_blockwise_graph(
        func, "z", "ik", "x", "ij", "y", "jk", numblocks={"x": (2, 2), "y": (2, 2)}
    )
    check_consistent_with_graph(block_fn, graph)


def test_make_blockwise_function_contract_1d():
    func = lambda x: 0

    block_fn = make_blockwise_function(
        func, "z", "j", "x", "ij", numblocks={"x": (1, 2)}
    )

    graph = make_blockwise_graph(func, "z", "j", "x", "ij", numblocks={"x": (1, 2)})
    check_consistent_with_graph(block_fn, graph)


def test_make_blockwise_function_contract_0d():
    func = lambda x: 0

    block_fn = make_blockwise_function(
        func, "z", "", "x", "ij", numblocks={"x": (2, 2)}
    )

    graph = make_blockwise_graph(func, "z", "", "x", "ij", numblocks={"x": (2, 2)})
    check_consistent_with_graph(block_fn, graph)


def check_consistent_with_graph(block_fn, graph):
    for k, v in graph.items():
        assert block_fn(k) == v



import numpy as np
import pytest
import zarr
from numpy.testing import assert_array_equal
from rechunker.executors.python import PythonPipelineExecutor

from cubed.primitive.rechunk import rechunk
from cubed.tests.utils import execute_pipeline


@pytest.fixture(scope="module", params=[PythonPipelineExecutor()])
def executor(request):
    return request.param


@pytest.mark.parametrize(
    "shape, source_chunks, max_mem, target_chunks, expected_required_mem, expected_num_tasks",
    [
        # one target chunk is made up of two source chunks
        (
            (4, 4),
            (1, 2),
            1000,
            (1, 4),
            1000,
            1,
        ),
        # only enough memory for one source/target chunk
        (
            (4, 4),
            (1, 4),
            4 * 8 * 4,  # elts x itemsize x copies
            (4, 1),
            4 * 8 * 4,  # elts x itemsize x copies
            8,
        ),
    ],
)
def test_rechunk(
    tmp_path,
    executor,
    shape,
    source_chunks,
    max_mem,
    target_chunks,
    expected_required_mem,
    expected_num_tasks,
):
    source = zarr.ones(shape, chunks=source_chunks, store=tmp_path / "source.zarr")
    max_mem = max_mem
    target_store = tmp_path / "target.zarr"
    temp_store = tmp_path / "temp.zarr"

    pipeline = rechunk(
        source,
        target_chunks=target_chunks,
        max_mem=max_mem,
        target_store=target_store,
        temp_store=temp_store,
    )

    assert pipeline.target_array.shape == shape
    assert pipeline.target_array.dtype == source.dtype
    assert pipeline.target_array.chunks == target_chunks

    assert pipeline.required_mem == expected_required_mem

    assert pipeline.num_tasks == expected_num_tasks

    execute_pipeline(pipeline, executor=executor)

    res = zarr.open(target_store)
    assert_array_equal(res[:], np.ones(shape))
    assert res.chunks == target_chunks


def test_rechunk_max_mem_exceeded(tmp_path):
    source = zarr.ones((4, 4), chunks=(2, 2), store=tmp_path / "source.zarr")
    max_mem = 16
    target_store = tmp_path / "target.zarr"
    temp_store = tmp_path / "temp.zarr"

    # max mem is reduced by a factor of 4 from 16 to 4
    with pytest.raises(
        ValueError, match=r"Source chunk memory \(32\) exceeds max_mem \(4\)"
    ):
        rechunk(
            source,
            target_chunks=(4, 1),
            max_mem=max_mem,
            target_store=target_store,
            temp_store=temp_store,
        )



import copy
import logging
import time
from functools import partial
from typing import Callable, Iterable

from lithops.executors import FunctionExecutor
from lithops.wait import ALWAYS, ANY_COMPLETED
from rechunker.types import ParallelPipelines, PipelineExecutor
from six import reraise

from cubed.core.array import TaskEndEvent
from cubed.core.plan import visit_nodes
from cubed.runtime.backup import should_launch_backup
from cubed.runtime.types import DagExecutor

logger = logging.getLogger(__name__)

# Lithops represents delayed execution tasks as functions that require
# a FunctionExecutor.
Task = Callable[[FunctionExecutor], None]


class LithopsPipelineExecutor(PipelineExecutor[Task]):
    """An execution engine based on Lithops."""

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def pipelines_to_plan(self, pipelines: ParallelPipelines) -> Task:
        tasks = []
        for pipeline in pipelines:
            stage_tasks = []
            for stage in pipeline.stages:
                if stage.mappable is not None:
                    stage_func = build_stage_mappable_func(stage, pipeline.config)
                    stage_tasks.append(stage_func)
                else:
                    stage_func = build_stage_func(stage, pipeline.config)
                    stage_tasks.append(stage_func)

            # Stages for a single pipeline must be executed in series
            tasks.append(partial(_execute_in_series, stage_tasks))

        return partial(_execute_in_series, tasks)

    def execute_plan(self, plan: Task, **kwargs):
        merged_kwargs = {**self.kwargs, **kwargs}
        with FunctionExecutor(**merged_kwargs) as executor:
            plan(executor)


def run_func(input, func=None, config=None, name=None):
    result = func(input, config=config)
    return result


def map_unordered(
    lithops_function_executor,
    map_function,
    map_iterdata,
    include_modules=[],
    max_failures=3,
    use_backups=False,
    return_stats=False,
    **kwargs,
):
    """
    Apply a function to items of an input list, yielding results as they are completed
    (which may be different to the input order).

    A generalisation of Lithops `map`, with retries, and relaxed return ordering.

    :param lithops_function_executor: The Lithops function executor to use.
    :param map_function: The function to map over the data.
    :param map_iterdata: An iterable of input data.
    :param include_modules: Modules to include.
    :param max_failures: The number of task failures to allow before raising an exception.
    :param use_backups: Whether to launch backup tasks to mitigate against slow-running tasks.
    :param return_stats: Whether to return lithops stats.

    :return: Function values (and optionally stats) as they are completed, not necessarily in the input order.
    """
    failures = 0
    return_when = ALWAYS if use_backups else ANY_COMPLETED

    inputs = map_iterdata
    tasks = {}
    start_times = {}
    end_times = {}
    backups = {}
    pending = []

    # can't use functools.partial here as we get an error in lithops
    # also, lithops extra_args doesn't work for this case
    partial_map_function = lambda x: map_function(x, **kwargs)

    futures = lithops_function_executor.map(
        partial_map_function, inputs, include_modules=include_modules
    )
    tasks.update({k: v for (k, v) in zip(futures, inputs)})
    start_times.update({k: time.monotonic() for k in futures})
    pending.extend(futures)

    while pending:
        finished, pending = lithops_function_executor.wait(
            pending, throw_except=False, return_when=return_when, show_progressbar=False
        )

        failed = []
        for future in finished:
            if future.error:
                failures += 1
                if failures > max_failures:
                    # re-raise exception
                    # TODO: why does calling status not raise the exception?
                    future.status(throw_except=True)
                    reraise(*future._exception)
                failed.append(future)
            else:
                end_times[future] = time.monotonic()
                if return_stats:
                    yield future.result(), future.stats
                else:
                    yield future.result()

            if use_backups:
                # remove backups
                backup = backups.get(future, None)
                if backup:
                    if backup in pending:
                        pending.remove(backup)
                    del backups[future]
                    del backups[backup]

        if failed:
            # rerun and add to pending
            inputs = [v for (fut, v) in tasks.items() if fut in failed]
            # TODO: de-duplicate code from above
            futures = lithops_function_executor.map(
                partial_map_function,
                inputs,
                include_modules=include_modules,
            )
            tasks.update({k: v for (k, v) in zip(futures, inputs)})
            start_times.update({k: time.monotonic() for k in futures})
            pending.extend(futures)

        if use_backups:
            now = time.monotonic()
            for future in copy.copy(pending):
                if future not in backups and should_launch_backup(
                    future, now, start_times, end_times
                ):
                    inputs = [v for (fut, v) in tasks.items() if fut == future]
                    logger.info("Running backup task for %s", inputs)
                    futures = lithops_function_executor.map(
                        partial_map_function,
                        inputs,
                        include_modules=include_modules,
                    )
                    tasks.update({k: v for (k, v) in zip(futures, inputs)})
                    start_times.update({k: time.monotonic() for k in futures})
                    pending.extend(futures)
                    pending.remove(future)  # throw away slow one
                    backup = futures[0]  # TODO: launch multiple backups at once
                    backups[future] = backup
                    backups[backup] = future
            time.sleep(1)


def execute_dag(dag, callbacks=None, array_names=None, **kwargs):
    use_backups = kwargs.pop("use_backups", False)
    with FunctionExecutor(**kwargs) as executor:
        for name, node in visit_nodes(dag):
            pipeline = node["pipeline"]
            for stage in pipeline.stages:
                if stage.mappable is not None:
                    for _, stats in map_unordered(
                        executor,
                        run_func,
                        list(stage.mappable),
                        func=stage.function,
                        config=pipeline.config,
                        name=name,
                        use_backups=use_backups,
                        return_stats=True,
                    ):
                        if callbacks is not None:
                            event = lithops_stats_to_task_end_event(name, stats)
                            [callback.on_task_end(event) for callback in callbacks]
                else:
                    raise NotImplementedError()


def lithops_stats_to_task_end_event(name, stats):
    return TaskEndEvent(
        array_name=name,
        task_create_tstamp=stats["host_job_create_tstamp"],
        function_start_tstamp=stats["worker_func_start_tstamp"],
        function_end_tstamp=stats["worker_func_end_tstamp"],
        task_result_tstamp=stats["host_status_done_tstamp"],
        peak_memory_start=stats["worker_peak_memory_start"],
        peak_memory_end=stats["worker_peak_memory_end"],
    )


def build_stage_mappable_func(
    stage, config, name=None, callbacks=None, use_backups=False
):
    def sf(mappable):
        return stage.function(mappable, config=config)

    def stage_func(lithops_function_executor):
        for _, stats in map_unordered(
            lithops_function_executor,
            sf,
            list(stage.mappable),
            use_backups=use_backups,
            return_stats=True,
        ):
            if callbacks is not None:
                event = lithops_stats_to_task_end_event(name, stats)
                [callback.on_task_end(event) for callback in callbacks]

    return stage_func


def build_stage_func(stage, config):
    def sf():
        return stage.function(config=config)

    def stage_func(lithops_function_executor):
        futures = lithops_function_executor.call_async(sf, ())
        lithops_function_executor.get_result(futures)

    return stage_func


def _execute_in_series(
    tasks: Iterable[Task], lithops_function_executor: FunctionExecutor
) -> None:
    for task in tasks:
        task(lithops_function_executor)


class LithopsDagExecutor(DagExecutor):
    """An execution engine that uses Lithops."""

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def execute_dag(self, dag, callbacks=None, array_names=None, **kwargs):
        merged_kwargs = {**self.kwargs, **kwargs}
        execute_dag(dag, callbacks=callbacks, array_names=array_names, **merged_kwargs)



# Use type code from numpy.array_api
from numpy.array_api._dtypes import (  # noqa: F401
    _boolean_dtypes,
    _dtype_categories,
    _floating_dtypes,
    _integer_dtypes,
    _integer_or_boolean_dtypes,
    _numeric_dtypes,
    bool,
    float32,
    float64,
    int8,
    int16,
    int32,
    int64,
    uint8,
    uint16,
    uint32,
    uint64,
)

_signed_integer_dtypes = (int8, int16, int32, int64)

_unsigned_integer_dtypes = (uint8, uint16, uint32, uint64)



import collections
import itertools
import platform
import sys
import sysconfig
import traceback
from dataclasses import dataclass
from math import prod
from pathlib import Path
from posixpath import join
from resource import RUSAGE_SELF, getrusage
from typing import Union
from urllib.parse import quote, unquote, urlsplit, urlunsplit

import numpy as np
from dask.array.core import _check_regular_chunks
from dask.utils import cached_cumsum

PathType = Union[str, Path]


def chunk_memory(dtype, chunksize):
    """Calculate the amount of memory in bytes that a single chunk uses."""
    return np.dtype(dtype).itemsize * prod(chunksize)


def get_item(chunks, idx):
    """Convert a chunk index to a tuple of slices."""
    starts = tuple(cached_cumsum(c, initial_zero=True) for c in chunks)
    loc = tuple((start[i], start[i + 1]) for i, start in zip(idx, starts))
    return tuple(slice(*s, None) for s in loc)


def join_path(dir_url: PathType, child_path: str) -> str:
    """Combine a URL for a directory with a child path"""
    parts = urlsplit(str(dir_url))
    new_path = quote(join(unquote(parts.path), child_path))
    split_parts = (parts.scheme, parts.netloc, new_path, parts.query, parts.fragment)
    return urlunsplit(split_parts)


def memory_repr(num):
    """Convert bytes to a human-readable string in decimal form.
    1 KB is 1,000 bytes, 1 MB is 1,000,000 bytes, and so on.

    Parameters
    ----------
    num: int
        Number of bytes

    Returns
    -------
    str
    """
    if num < 1000.0:
        return f"{num} bytes"
    num /= 1000.0
    for x in ["KB", "MB", "GB", "TB", "PB"]:
        if num < 1000.0:
            return f"{num:3.1f} {x}"
        num /= 1000.0


def peak_memory():
    """Return the peak memory usage in bytes."""
    ru_maxrss = getrusage(RUSAGE_SELF).ru_maxrss
    # note that on Linux ru_maxrss is in KiB, while on Mac it is in bytes
    # see https://pythonspeed.com/articles/estimating-memory-usage/#measuring-peak-memory-usage
    return ru_maxrss * 1024 if platform.system() == "Linux" else ru_maxrss


def to_chunksize(chunkset):
    """Convert a chunkset to a chunk size for Zarr.

    Parameters
    ----------
    chunkset: tuple of tuples of ints
        From the ``.chunks`` attribute of an ``Array``

    Returns
    -------
    Tuple of ints
    """

    if not _check_regular_chunks(chunkset):
        raise ValueError("Array must have regular chunks")

    return tuple(c[0] for c in chunkset)


@dataclass
class StackSummary:
    """Like Python's ``FrameSummary``, but with module information."""

    filename: str
    lineno: int
    name: str
    module: str

    def is_cubed(self):
        """Return True if this stack frame is a Cubed call."""
        return self.module.startswith("cubed.") and not self.module.startswith(
            "cubed.tests."
        )

    def is_on_python_lib_path(self):
        """Return True if this stack frame is from a library on Python's library path."""
        python_lib_path = sysconfig.get_path("purelib")

        return self.filename.startswith(python_lib_path)


def extract_stack_summaries(frame, limit=None):
    """Like Python's ``StackSummary.extract``, but returns module information."""
    frame_gen = traceback.walk_stack(frame)

    # from StackSummary.extract
    if limit is None:
        limit = getattr(sys, "tracebacklimit", None)
        if limit is not None and limit < 0:
            limit = 0
    if limit is not None:
        if limit >= 0:
            frame_gen = itertools.islice(frame_gen, limit)
        else:
            frame_gen = collections.deque(frame_gen, maxlen=-limit)
    # end from StackSummary.extract

    stack_summaries = []
    for f, _ in frame_gen:
        module = f.f_globals.get("__name__", "")
        summary = StackSummary(
            filename=f.f_code.co_filename,
            lineno=f.f_lineno,
            name=f.f_code.co_name,
            module=module,
        )
        stack_summaries.append(summary)
    stack_summaries.reverse()

    return stack_summaries
